/* $Id: comm.c,v 1.35 2007-05-24 16:08:55 kristaps Exp $ */
/*
Copyright (c) 2007 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/param.h>

#include <assert.h>
#include <err.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <libgen.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <curl/curl.h>

#include "nlpcrawl.h"
#include "comm.h"


#define COMM_TIMEO_DNS	(-1)
#define COMM_TIMEO_CONN	(60)

/*
 * Shared handle by all easy handles.  Should stay active for the full
 * system run-time.
 */
static CURLSH		*share = NULL;

/*
 * Multi-comm handle used for the set of currently-persisting
 * connections.  This should stay active for the full run-time.
 */
static CURLM		*cookie = NULL;

/*
 * Call-back for header data.
 */
static size_t (*write_head)(void *, size_t, size_t, void *);

/*
 * Call-back for page body data.
 */
static size_t (*write_data)(void *, size_t, size_t, void *);

/*
 * Length of persistent header buffers.
 */
#define COMM_HDR_LEN	(256)

/*
 * Header containers for persistent, un-changing headers: language
 * (Accept-Language), character set (Accept-Charset) and User-Agent.
 */
static char		head_lang[COMM_HDR_LEN];
static char		head_charset[COMM_HDR_LEN];
static char		head_agent[COMM_HDR_LEN];


static void
comm_stats_double(struct comm *p, CURLINFO opt, double *val)
{
	CURLcode	ret;
	CURL		*curl;

	assert(NULL != val);
	assert(NULL != p);
	assert(NULL != p->handle);

	curl = (CURL *)p->handle;
	if (CURLE_OK == (ret = curl_easy_getinfo(curl, opt, val)))
		return;

	diex("%s: fail: %s, %s", __func__,
			curl_easy_strerror(ret), p->pretty);
	/*NOTREACHED*/
}


static void
comm_stats_fill(struct comm *p, struct comm_stats *s)
{
	double		d;

	assert(NULL != p);
	assert(NULL != s);

	comm_stats_double(p, CURLINFO_TOTAL_TIME, &d);
	s->ttime = (uint64_t)d;
	comm_stats_double(p, CURLINFO_NAMELOOKUP_TIME, &d);
	s->nl_time = (uint64_t)d;
	comm_stats_double(p, CURLINFO_CONNECT_TIME, &d);
	s->ct_time = (uint64_t)d;
	comm_stats_double(p, CURLINFO_PRETRANSFER_TIME, &d);
	s->pt_time = (uint64_t)d;
	comm_stats_double(p, CURLINFO_STARTTRANSFER_TIME, &d);
	s->st_time = (uint64_t)d;
	comm_stats_double(p, CURLINFO_SIZE_DOWNLOAD, &d);
	s->dl_size = (uint64_t)d;
}


static void
comm_opt_long(struct comm *p, CURLoption opt, long val)
{
	CURLcode	ret;
	assert(NULL != p);
	assert(NULL != p->handle);

	ret = curl_easy_setopt((CURL *)p->handle, opt, val);
	if (CURLE_OK == ret)
		return;
	diex("%s: fail: %s, %s", __func__,
			curl_easy_strerror(ret), p->pretty);
	/*NOTREACHED*/
}


static void
comm_opt_void(struct comm *p, CURLoption opt, void *val)
{
	CURLcode	ret;
	assert(NULL != p);
	assert(NULL != p->handle);
	assert(NULL != val);

	ret = curl_easy_setopt((CURL *)p->handle, opt, val);
	if (CURLE_OK == ret)
		return;
	diex("%s: fail: %s, %s", __func__,
			curl_easy_strerror(ret), p->pretty);
	/*NOTREACHED*/
}


static void
comm_hdr_add(struct comm *p, char field[COMM_HEAD_SZ], 
		const char *key, const char *val)
{
	assert(0 == field[0]);
	assert(NULL != p);
	assert(NULL != key);
	assert(NULL != val);

	strncpy(field, key, COMM_HEAD_SZ - 1);
	assert(0 == field[COMM_HEAD_SZ - 1]);

	strncat(field, val, COMM_HEAD_SZ - strlen(field) - 1);
	assert(0 == field[COMM_HEAD_SZ - 1]);

	strncat(field, "\r\n", COMM_HEAD_SZ - strlen(field) - 1);
	assert(0 == field[COMM_HEAD_SZ - 1]);

	p->head.handle = curl_slist_append(p->head.handle, field);
	assert(NULL != p->head.handle);

}


static void
comm_hdr_init(char field[COMM_HEAD_SZ], const char *key, char *val)
{
	char		*t, *tt;

	assert(0 == field[0]);
	assert(NULL != key);
	assert(NULL != val);

	if (COMM_HEAD_SZ < strlen(val) + strlen(key) + 10)
		diex("%s: fail: header overrun: %s", 
				__func__, key);

	while (NULL != (t = strsep(&val, " "))) { 
		if (0 == strlen(t))
			continue;
		for (tt = t; 0 != *tt; tt++) 
			if (isalpha(*tt))
				*tt = tolower(*tt);
		strncat(field, t, COMM_HEAD_SZ - strlen(field) - 1);
		strncat(field, ",", COMM_HEAD_SZ - strlen(field) - 1);
	} 
	strncat(field, "*\r\n" , COMM_HEAD_SZ - strlen(field) - 1);
	assert(0 == field[COMM_HEAD_SZ - 1]);
}


void
comm_hdr_lang(char *val)
{
	comm_hdr_init(head_lang, "Accept-Language", val);
}


void
comm_hdr_charset(char *val)
{
	comm_hdr_init(head_charset, "Accept-Content", val);
}


void
comm_hdr_agent(const char *val)
{
	assert(NULL != val);
	assert(0 == head_agent[0]);

	if (COMM_HEAD_SZ < strlen(val) + strlen("User-Agent: ") + 1)
		diex("%s: fail: header overrun: %s", __func__, val);

	strncat(head_agent, val, COMM_HEAD_SZ - strlen(head_agent) - 1);
	assert(0 == head_agent[COMM_HEAD_SZ - 1]);

	strncat(head_agent, "\r\n", 
			COMM_HEAD_SZ - strlen(head_agent) - 1);
	assert(0 == head_agent[COMM_HEAD_SZ - 1]);
}


void
comm_free(struct comm *p)
{
	assert(NULL != p);
	assert(NULL != p->handle);

	curl_easy_cleanup((CURL *)p->handle);
	p->handle = NULL;
}


void
comm_close(struct comm *p, struct comm_stats *s)
{
	CURLMcode	ret;
	CURL		*curl;

	assert(NULL != p);
	assert(NULL != p->handle);
	assert(NULL != cookie);
	assert(NULL != s);
	assert(NULL != p->hash);
	assert(-1 != p->fd);

	curl = (CURL *)p->handle;

	if (-1 == close(p->fd))
		die("%s: fail: close: %s", __func__, p->pretty);
	p->fd = -1;

	assert(NULL != p->head.handle);
	curl_slist_free_all((struct curl_slist *)p->head.handle);
	
	md5_hash_close(&p->hash, p->pair->record.data);

	comm_stats_fill(p, s);

	if (CURLM_OK == (ret = curl_multi_remove_handle(cookie, curl)))
		return;

	diex("%s: fail: %s, %s", __func__,
			curl_multi_strerror(ret), p->pretty);
	/*NOTREACHED*/
}


void
comm_new(struct comm *p)
{
	assert(NULL != p);
	p->fd = -1;
	if (NULL == (p->handle = curl_easy_init()))
		diex("fail: %s: curl_easy_init", __func__);
}


void
comm_open(struct comm *p, const char *prefix, struct nlpcdbrp *pair)
{
	CURLMcode	ret;
	CURL		*curl;
	struct nlpcdbr	*r;
	char		path[MAXPATHLEN], dir[MAXPATHLEN], *dp, *t;
	char		dt[32];
	time_t		tm;

	assert(NULL != prefix);
	assert(NULL != pair);
	assert(NULL != p);
	assert(NULL != p->handle);
	assert(NULL == p->hash);
	assert(NULL == p->pair);
	assert(0 == p->status);
	assert(-1 == p->fd);
	assert(NULL == p->head.handle);

	p->pair = pair;
	r = &p->pair->record;
	curl = (CURL *)p->handle;

	/* 
	 * Set up transfer-time-persistent buffers.
	 */

	assert(0 == p->uri[0]);
	strncpy(p->uri, uri_tostring(&r->uri), NLPCR_URI_MAXSZ);
	assert(0 == p->uri[NLPCR_URI_MAXSZ - 1]);
	assert(0 == p->pretty[0]);
	strncpy(p->pretty, uri_topstring(&r->uri), NLPCR_URI_MAXSZ);
	assert(0 == p->pretty[NLPCR_URI_MAXSZ - 1]);

	/* 
	 * Set up persistent headers. 
	 */
	
	if (0 != r->ftime) {
		tm = (time_t)r->ftime;
		memcpy(dt, ctime(&tm), 32);
		dt[strlen(dt) - 1] = 0;
		comm_hdr_add(p, p->head.date, 
				"If-Modified-Since: ", dt);
	}

	comm_hdr_add(p, p->head.agent, "User-Agent: ", head_agent);

	assert(0 == p->head.lang[0]);
	if (0 != head_lang[0])
		comm_hdr_add(p, p->head.lang, "Accept-Language: ",
				head_lang);

	assert(0 == p->head.charset[0]);
	if (0 != head_lang[0])
		comm_hdr_add(p, p->head.charset, "Accept-Content: ",
				head_charset);

	/* 
	 * Prepare curl transfer object.
	 */

	assert(NULL != share);

	comm_opt_void(p, CURLOPT_URL, p->uri);
	comm_opt_void(p, CURLOPT_WRITEDATA, p);
	comm_opt_void(p, CURLOPT_HEADERDATA, p);
	comm_opt_void(p, CURLOPT_HEADERFUNCTION, write_head);
	comm_opt_void(p, CURLOPT_WRITEFUNCTION, write_data);
	comm_opt_void(p, CURLOPT_HTTPHEADER, p->head.handle);
	comm_opt_void(p, CURLOPT_SHARE, share);

	comm_opt_long(p, CURLOPT_FOLLOWLOCATION, 0);
	comm_opt_long(p, CURLOPT_TIMEOUT, COMM_TIMEO_CONN);
	comm_opt_long(p, CURLOPT_NOSIGNAL, 1);
	comm_opt_long(p, CURLOPT_DNS_CACHE_TIMEOUT, COMM_TIMEO_DNS);

	assert(NULL != cookie);
	assert(NULL != curl);

	if (CURLM_OK != (ret = curl_multi_add_handle(cookie, curl)))
		diex("%s: fail: %s, %s", __func__,
				curl_multi_strerror(ret), p->pretty);

	/*
	 * Configure the cache target.  Note that we only prime the
	 * transfer hash if the file's being written (to avoid erroneous
	 * duplicate detections).
	 */

	md5_hash_open(&p->hash);

	memset(path, 0, MAXPATHLEN);
	memset(dir, 0, MAXPATHLEN);

	t = nlpcdbr_hash2path(prefix, pair->hash);

	memcpy(path, t, MAXPATHLEN);
	memcpy(dir, path, MAXPATHLEN);

	if (NULL == (dp = dirname(dir)))
		die("%s: fail: dirname: %s", __func__, path);
	if (mkdir_p(dp, 0755))
		die("%s: fail: mkdir_p: %s", __func__, dp);

	p->fd = open(path, O_WRONLY | O_TRUNC | O_CREAT, 0644);
	if (-1 == p->fd) 
		die("%s: fail: open: %s", __func__, path);
}


void
comm_init(size_t (*head)(void *, size_t, size_t, void *),
		size_t (*data)(void *, size_t, size_t, void *))
{
	CURLcode 	cc;
	CURLSHcode	cs;
	CURLMcode	cm;

	assert(NULL != head);
	assert(NULL != data);

	write_head = head;
	write_data = data;

	cc = curl_global_init(CURL_GLOBAL_ALL);
	if (CURLE_OK != cc)
		diex("%s: fail: curl-global-init: %s", __func__,
				curl_easy_strerror(cc));

	assert(NULL == cookie);
	if (NULL == (cookie = curl_multi_init()))
		diex("%s: fail: multi-curl-init", __func__);

#ifdef CURLMOPT_PIPELININE
	cm = curl_multi_setopt(cookie, CURLMOPT_PIPELINING, 1);
	if (CURLM_OK != cm)
		diex("%s: fail: multi-curl-setopt: %s", __func__,
				curl_multi_strerror(cm));
#else
	(void)cm;
#endif

	if (NULL == (share = curl_share_init()))
		diex("%s: fail: share-curl-init", __func__);
	cs = curl_share_setopt(share, CURLSHOPT_SHARE, 
			CURL_LOCK_DATA_DNS);
	if (CURLSHE_OK != cs)
		diex("%s: fail: share-curl-setopt", __func__);

	cs = curl_share_setopt(share, CURLSHOPT_SHARE,
			CURL_LOCK_DATA_COOKIE);
	if (CURLSHE_OK != cs)
		diex("%s: fail: share-curl-setopt", __func__);
}


void
comm_teardown(void)
{
	CURLMcode 	mc;
	CURLSHcode 	ms;

	assert(NULL != cookie);
	if (CURLM_OK != (mc = curl_multi_cleanup(cookie))) 
		diex("%s: fail: multi-curl-cleanup: %s",
				__func__, curl_multi_strerror(mc));
	cookie = NULL;

	assert(NULL != share);
	if (CURLSHE_OK != (ms = curl_share_cleanup(share)))
		diex("%s: fail: share-curl-cleanup: %s",
				__func__, curl_share_strerror(ms));
	share = NULL;

	curl_global_cleanup();
}


void *
comm_report_next(int *run, int *type)
{
	CURLMsg		*msg;

	assert(NULL != cookie);
	assert(NULL != run);
	assert(NULL != type);

	if (NULL == (msg = curl_multi_info_read(cookie, run)))
		return(NULL);

	switch(msg->data.result) {
	case (CURLE_OK):
		*type = COMM_OK;
		break;
	case (CURLE_READ_ERROR):
	case (CURLE_WRITE_ERROR):
		*type = COMM_CALLBACK;
		break;
	case (CURLE_FTP_CANT_GET_HOST):
	case (CURLE_COULDNT_RESOLVE_PROXY):
	case (CURLE_COULDNT_RESOLVE_HOST):
		*type = COMM_RESOLVE;
		break;
	case (CURLE_FTP_CANT_RECONNECT):
	case (CURLE_COULDNT_CONNECT):
		*type = COMM_CONNECT;
		break;
	case (CURLE_OPERATION_TIMEDOUT):
		*type = COMM_TIMEOUT;
		break;
	case (CURLE_HTTP_RETURNED_ERROR):
		*type = COMM_SERVER;
		break;
	case (CURLE_URL_MALFORMAT):
		*type = COMM_SYNTAX;
		break;
	case (CURLE_OUT_OF_MEMORY):
	case (CURLE_ABORTED_BY_CALLBACK):
	case (CURLE_BAD_FUNCTION_ARGUMENT):
	case (CURLE_FILESIZE_EXCEEDED):
		*type = COMM_INTERNAL;
		break;
	default:
		*type = COMM_GENERAL;
		break;
	}
	return(msg->easy_handle);
}


int
comm_perform(void)
{
	CURLMcode	cm;
	int		run;

	assert(NULL != cookie);

	cm = CURLM_CALL_MULTI_PERFORM;
	for (;;) if (CURLM_CALL_MULTI_PERFORM != 
			(cm = curl_multi_perform(cookie, &run)))
		break;

	if (CURLM_OK == cm)
		return(run);

	diex("%s: fail: curl-multi-perform: %s", 
			__func__, curl_multi_strerror(cm));
	/* NOTREACHED */
}


void
comm_wait(void)
{
	fd_set		set;
	CURLMcode	cm;
	int		mfds;
	fd_set		r1, r2;
	struct timeval	tm;

	assert(NULL != cookie);

	FD_ZERO(&set);
	FD_ZERO(&r1);
	FD_ZERO(&r2);

	cm = curl_multi_fdset(cookie, &set, &r1, &r2, &mfds);
	if (CURLM_OK != cm)
		diex("%s: fail: multi-fdset: %s", __func__,
				curl_multi_strerror(cm));
	if (-1 == mfds)
		return;

	timerclear(&tm);
	tm.tv_sec = 0;
	tm.tv_usec = 100000;

	if (-1 == select(mfds, &set, NULL, NULL, &tm))
		die("%s: fail: select", __func__);
}
