/* $Id: comm.h,v 1.14 2007-05-23 15:04:28 kristaps Exp $ */
/*
Copyright (c) 2007 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef COMM_H
#define COMM_H

__BEGIN_DECLS

/* 
 * Length of header line (inclusive trailer space).
 */
#define COMM_HEAD_SZ	(256)

struct comm {
	char		uri[NLPCR_URI_MAXSZ]; /* Raw URI form.       */
	char		pretty[NLPCR_URI_MAXSZ]; /* Pretty URI.      */
	int		fd;		/* Open file-descriptor.     */
	uint32_t	status;		/* Transfer status-code.     */
	struct nlpcdbrp	*pair;		/* Relevant record.          */
	void		*handle;	/* Underlying comm handle.   */
	void		*hash;		/* Hashing session.          */

	struct {
		void	*handle;	/* Header-field handle.      */
		char	date[COMM_HEAD_SZ]; /* If-modified-since.    */
		char	agent[COMM_HEAD_SZ]; /* User-agent.          */
		char	lang[COMM_HEAD_SZ]; /* Accept-language.      */
		char	charset[COMM_HEAD_SZ]; /* Accept-charset.    */
	} head;
};

struct comm_stats {
	uint64_t	ttime;		/* Total transfer time.      */
	uint64_t	nl_time;	/* Namelookup time.          */
	uint64_t	ct_time;	/* Connect time.             */
	uint64_t	pt_time;	/* Pre-transfer time.        */
	uint64_t	st_time;	/* Start-transfer time.      */
	uint64_t	dl_size;	/* Download size.            */
};


/*
 * Reset a comm handle between transfer sessions.
 */
#define comm_reset(p) do {                                           \
		void *sv = (p)->handle;                              \
		memset((p), 0, sizeof(struct comm));                 \
		(p)->fd = -1;                                        \
		(p)->handle = sv;                                    \
	} while(0)


/*
 * Install a space-separated list of acceptable languages (this will be
 * terminated with the wild-card).
 */
void
comm_hdr_lang(char *val);


/*
 * Install a space-separated list of acceptable charsets (this will be
 * terminated with the wild-card).
 */
void
comm_hdr_charset(char *val);


/*
 * Install the user-agent.
 */
void
comm_hdr_agent(const char *val);


/*
 * Free up the memory held by a communicator.
 */
void
comm_free(struct comm *p);


/*
 * Pre-initialise a communication object.
 */
void
comm_new(struct comm *p);


/*
 * Close a transfer opened with comm_open.
 */
void
comm_close(struct comm *p, struct comm_stats *s);


/*
 * Open a transfer to an endpoint.  Also, if not read-only, prepares the
 * local cache for writing.
 */
void
comm_open(struct comm *p, const char *prefix, struct nlpcdbrp *pair);


/*
 * Initialise the communication framework.  Any comm functions (beyond
 * the comm_hdr family) must follow this.  This should be symmetric to
 * comm_teardown.
 */
void
comm_init(size_t (*head)(void *, size_t, size_t, void *),
		size_t (*data)(void *, size_t, size_t, void *));


/*
 * Teardown the communication subsystem.  May only be called after
 * comm_init.
 */
void
comm_teardown(void);


#define COMM_OK		(0)
#define COMM_CALLBACK	(1)
#define COMM_RESOLVE	(2)
#define COMM_CONNECT	(3)
#define COMM_TIMEOUT	(4)
#define COMM_SERVER	(5)
#define COMM_SYNTAX	(6)
#define COMM_INTERNAL	(7)
#define COMM_GENERAL	(8)

/*
 * Report on a completed or aborted transfer.  Sets "type" to be the
 * general status code of the transfer.
 */
void *
comm_report_next(int *cookie, int *type);


/*
 * Perform work and scan for completed transfers.  Returns an
 * indentifier for the completed (or aborted) transfer.  This cookie can
 * be passed into comm_report_next in order to get further data.
 */
int
comm_perform(void);


/*
 * Wait for data on any of the comm descriptors.  This internally uses
 * select.  FIXME - use a more efficient means of polling (like poll).
 */
void
comm_wait(void);


__END_DECLS

#endif /*!COMM_H*/
