/* $Id: fetch.c,v 1.137 2007-05-25 11:47:12 kristaps Exp $ */
/*
Copyright (c) 2007 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/param.h>

#include <assert.h>
#include <err.h>
#include <errno.h>
#include <getopt.h>
#include <inttypes.h>
#include <pthread.h>
#include <signal.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <curl/curl.h>

#include "nlpcrawl.h"
#include "comm.h"


/* 
 * Default number of active communication descriptors.  This refers to
 * the number of simultaneous transfers; the shell's ulimit must be able
 * to handle this value.
 */
#define FETCH_DEF_COMM	(512)

/*
 * Number of elements to hosts to allow to queue. It makes sense to have
 * this larger than the number of communication objects.
 */
#define FETCH_QUEUE_SZ	(1024)

/*
 * Number of connections (maximum) to simultaneously open per host.
 */
#ifdef CURLMOPT_PIPELINING
# define FETCH_QUEUE_BURST_SZ (6)
#else
# define FETCH_QUEUE_BURST_SZ (2)
#endif

/*
 * Maximum number of per-host fetch objects to queue.  This should be a
 * fairly large number: however, if too large, then a single entry can
 * dominate the connection time at (this * burst_wait) possible
 * overhead.
 */
#ifdef CURLMOPT_PIPELINING
# define FETCH_QUEUE_BURST_MAX (30)
#else
# define FETCH_QUEUE_BURST_MAX (10)
#endif

/*
 * Politeness policy wait delay (between bursts) as offset by the last
 * transfer sessions's end time.
 */
#define FETCH_QUEUE_BURST_WAIT (10)


struct host_t 
{
	void		*queue;		/* Queue of perhost nlpcdbrp. */
	size_t		queue_sz;	/* Count in queue.            */
	time_t		last;		/* Last fetch from host.      */
	size_t		pending;	/* Number of pending fetches. */
	int		valid;
};


static struct opt_t 
{
	size_t		ptrigger;	/* Fetch queue fill-trigger. */
	size_t		pburst;		/* Fetch burst size.         */
	size_t		pwidth;		/* Fetch queue width.        */
	size_t		pdepth;		/* Fetch queue depth.        */
	size_t		pwait;		/* Fetch wait seconds.       */
	size_t		ppar;		/* Fetch parallel comms.     */
	void		*wqueue;	/* Write queue.              */
	void		*rqueue;	/* Read queue.               */
	char		*cache;		/* Cache file directory.     */
	int		ascend;		/* Perform path-ascendence.  */
	void		*enc;		/* Dict of encodings.        */
	void		*lang;		/* Dict of languages.        */
	void		*auth;
	void		*scheme;
} opts;


static int
comm_content_lang(struct http_header *h, struct comm *c)
{
	struct nlpcdbr	*r;

	assert(NULL != h);
	assert(NULL != c);
	assert(HTTP_CONTENT_LANG == h->type);

	r = &c->pair->record;
	memset(r->lang, 0, NLPCR_CHARSETLEN);

	if (NULL != h->data.clang) 
		strncpy(r->lang, h->data.clang, NLPCR_LANGLEN);

	if (0 != r->lang[NLPCR_LANGLEN - 1]) {
		ddbg("%s: warn: language overrun: %s", 
				__func__, c->pretty);
		r->state = NLPCDBR_UNREACH;
		r->code = NLPCDBR_SYNTAX;
		return(0);
	} 

	if ( ! dict_has(&opts.lang, r->lang, strlen(r->lang))) {
		ddbg("%s: warn: non-matching language: [%s], %s", 
				__func__, r->lang, c->pretty);
		r->state = NLPCDBR_UNREACH;
		r->code = NLPCDBR_LANG;
		return(0);
	}

	ddbg("%s: accept: content-language: [%s], %s", __func__,
			r->lang, c->pretty);
	return(1);
}


static int
comm_content_type(struct http_header *h, struct comm *c)
{
	struct nlpcdbr	*r;

	assert(NULL != h);
	assert(NULL != c);
	assert(HTTP_CONTENT_TYPE == h->type);

	r = &c->pair->record;

	if (NLPCDBR_TYPE_NONE == (r->type = h->data.ctype.media)) {
		ddbg("%s: warn: unknown media: %s", 
				__func__, c->pretty);
		r->state = NLPCDBR_UNREACH;
		r->code = NLPCDBR_CTYPE;
		return(0);
	}

	memset(r->charset, 0, NLPCR_CHARSETLEN);
	if (NULL != h->data.ctype.charset) 
		strncpy(r->charset, h->data.ctype.charset,
				NLPCR_CHARSETLEN);

	if (0 != r->charset[NLPCR_CHARSETLEN - 1]) {
		ddbg("%s: warn: charset overrun: [%s], %s", __func__, 
				h->data.ctype.charset, c->pretty);
		r->state = NLPCDBR_UNREACH;
		r->code = NLPCDBR_SYNTAX;
		return(0);
	}

	if ( ! dict_has(&opts.enc, r->charset, strlen(r->charset))) {
		ddbg("%s: warn: non-matching charset: [%s], %s", 
				__func__, r->charset, c->pretty);
		r->state = NLPCDBR_UNREACH;
		r->code = NLPCDBR_CHARSET;
		return(0);
	}

	ddbg("%s: accept: content-type: [%s][%s], %s", __func__,
			nlpcdbr_type2str(r->type),
			r->charset, c->pretty);
	return(1);
}


static void
comm_enqueue(const struct uri *addr)
{
	struct nlpcdbrp	*n;
	assert(NULL != addr);

	calloc_s(n, 1, sizeof(struct nlpcdbrp));
	nlpcr_new(addr, n);
	squeue_add(opts.wqueue, n, SQUEUE_NOTIFY | SQUEUE_WAIT);
}


size_t
comm_head(void *ptr, size_t smem, size_t nmem, void *arg)
{
	struct http_header h;
	struct comm 	*co;
	struct nlpcdbr	*r;

	assert(NULL != arg);
	assert(NULL != ptr);
	assert(1 == smem);

	co = (struct comm *)arg;
	r = &co->pair->record;

#define mod(st, cd) do {                                              \
	co->pair->record.state = (st);                                \
	co->pair->record.code = (cd); } while(0)

#define mod_ret(st, cd) do {                                          \
	mod((st), (cd));                                              \
	return(nmem + 1);                                             \
	/* NOTREACHED */ } while(0)

	if (0 == co->status) {
		if ( ! http_status((char *)ptr, nmem, &co->status)) {
			ddbg("%s: malformed status line: %s", 
					__func__, co->pretty);
			mod_ret(NLPCDBR_UNREACH, NLPCDBR_SYNTAX);
		}
		switch(co->status) {
		case (200):
			break;
		case (304):
			mod_ret(NLPCDBR_FRESH, 0);
			/* NOTREACHED */
		case (201):
			/* FALLTHROUGH */
		case (300):
			/* FALLTHROUGH */
		case (302):
			/* FALLTHROUGH */
		case (303):
			/* FALLTHROUGH */
		case (307):
			mod(NLPCDBR_UNREACH, NLPCDBR_MOVED);
			break;
		case (301):
			mod(NLPCDBR_DEAD, NLPCDBR_MOVED);
			break;
		default:
			mod_ret(NLPCDBR_UNREACH, NLPCDBR_SERVER);
			/* NOTREACHED */
		}
		return(nmem);
	}

	if ( ! http_parse((char *)ptr, nmem, &h))
		return(nmem);

	switch (h.type) {
	case (HTTP_EXPIRES):
		r->expires = 
			(uint64_t)curl_getdate(h.data.expires, NULL);
		break;
	case (HTTP_PRAGMA):
		if (HTTP_PRAGMA_NOCACHE == h.data.pragma.pragma_type)
			if (h.data.pragma.data.nocache)
				r->expires = UINT64_MAX;
		break;
	case (HTTP_CONTENT_TYPE):
		if ( ! comm_content_type(&h, co))
			return(nmem + 1);
		break;
	case (HTTP_CONTENT_LANG):
		if ( ! comm_content_lang(&h, co))
			return(nmem + 1);
		break;
	case (HTTP_LOCATION):
		if ( ! opts.ascend) 
			comm_enqueue(&h.data.location);
		else while (uri_ascend(&h.data.location)) 
			comm_enqueue(&h.data.location);
		return(nmem + 1);
	default:
		break;
	}

	return(nmem);
}


size_t
comm_data(void *ptr, size_t smem, size_t nmem, void *arg)
{
	ssize_t		rc;
	struct comm	*p;

	assert(NULL != arg);
	assert(NULL != ptr);
	assert(1 == smem);

	p = (struct comm *)arg;

	assert(NULL != p->hash);
	md5_hash_update((const char *)ptr, nmem, p->hash);

	if (-1 == (rc = write(p->fd, ptr, nmem)))
		die("write: %s", p->pretty);
	if (rc != (ssize_t)nmem) 
		diex("write (underrun): %s", p->pretty);

	p->pair->record.size += nmem;

	return(nmem);
}


static void
thread_exec_fill(void *htab, void *queue, struct host_t *buf, 
		size_t *width, size_t *slots)
{
	struct nlpcdbrp	*p;
	struct host_t	*h;
	char		*auth;
	size_t		i;

	assert(NULL != queue);
	assert(NULL != htab);
	assert(NULL != buf);
	assert(NULL != width);
	assert(NULL != slots);

	assert(0 < opts.pdepth);

	for (p = NULL; ; ) {
		if (NULL != p)
			free(p);
		p = squeue_rem(opts.rqueue, SQUEUE_NOTIFY);
		if (NULL == p)
			break;
		auth = p->record.uri.auth;

		if ( ! hash_get(&htab, auth, NLPCR_URI_AUTHSZ, 
					(void **)&h)) {
			if (*width >= opts.pwidth)
				continue;
			for (i = 0; i < opts.pwidth; i++)
				if ( ! buf[i].valid)
					break;
			assert(i < opts.pwidth);
			h = &buf[i];

			h->valid = 1;
			h->pending = 0;
			h->queue_sz = 0;
			h->last = 0;
			assert(NULL != h->queue);

			q_add(&queue, h);
			hash_put(&htab, auth, NLPCR_URI_AUTHSZ, h);
			++(*width);
		}
	
		if (h->queue_sz >= opts.pdepth)
			continue;

		q_add(&h->queue, p);
		++(h->queue_sz);
		++(*slots);
		p = NULL;
		if (*slots >= (opts.pdepth * opts.pwidth))
			break;
	}
}


static void
thread_exec_rm(struct comm *p, int type)
{
	uint64_t	tm;
	struct nlpcdbr	*r;
	struct comm_stats stat;

	assert(NULL != p);
	assert(NULL != p->pair);

	tm = time_uint64();
	r = &p->pair->record;

	p->pair->record.atime = tm;
	p->pair->record.mtime = tm;
	if (COMM_OK == type)
		p->pair->record.ftime = tm;

#define set_state(st, cd) r->state = (st); r->code = (cd)

	switch (type) {
	case (COMM_OK):
		set_state(NLPCDBR_FRESH, NLPCDBR_OK);
		break;
	case (COMM_CALLBACK):
		break;
	case (COMM_RESOLVE):
		set_state(NLPCDBR_UNREACH, NLPCDBR_RESOLV);
		break;
	case (COMM_CONNECT):
		set_state(NLPCDBR_UNREACH, NLPCDBR_CONNECT);
		break;
	case (COMM_TIMEOUT):
		set_state(NLPCDBR_UNREACH, NLPCDBR_TIMEO);
		break;
	case (COMM_SERVER):
		set_state(NLPCDBR_UNREACH, NLPCDBR_SERVER);
		break;
	case (COMM_SYNTAX):
		set_state(NLPCDBR_DEAD, NLPCDBR_URL);
		break;
	case (COMM_INTERNAL):
		diex("fail: internal: %s", p->pretty);
		/* NOTREACHED */
	default:
		set_state(NLPCDBR_UNREACH, NLPCDBR_GENERAL);
		break;
	}

	comm_close(p, &stat);
	/*dbg("%s: adding to queue: %s", __func__, 
			p->pair->record.uri.auth);*/
	squeue_add(opts.wqueue, p->pair, SQUEUE_NOTIFY | SQUEUE_WAIT);
	comm_reset(p);
}


void
thread_exec_cleanup_comms(void *p)
{
	size_t		i;

	assert(NULL != p);
	for (i = 0; i < opts.ppar; i++)
		comm_free(&((struct comm *)p)[i]);
	free(p);
}


void
thread_exec_cleanup_hosts(void *p)
{
	size_t		i;

	assert(NULL != p);
	for (i = 0; i < opts.pwidth; i++)
		q_free(&((struct host_t *)p)[i].queue);
	free(p);
}


void *
thread_exec(void *arg)
{
	size_t		i, slots, width, pend;
	void		*hostsh, *handle, *commsq, *hostsq, *cbh;
	struct host_t	*hosts, *h;
	struct comm	*comms, *c;
	struct nlpcdbrp	*p;
	time_t		cur;
	int		type, cookie;
	char		*auth, addr[NLPCR_URI_MAXSZ + 10];

	assert(NULL == arg);

	comm_init(comm_head, comm_data);

	calloc_s(comms, opts.ppar, sizeof(struct comm));
	calloc_s(hosts, opts.pwidth, sizeof(struct host_t));

	q_new(&hostsq);
	q_new(&commsq);

	hash_new(&hostsh);
	hash_new(&cbh);

	for (i = 0; i < opts.ppar; i++) {
		comm_new(&comms[i]);
		q_add(&commsq, &comms[i]);
		hash_put(&cbh, &comms[i].handle, sizeof(void *),
				&comms[i]);
	}

	for (i = 0; i < opts.pwidth; i++)
		q_new(&hosts[i].queue);

	pthread_cleanup_push((void (*)(void *))comm_teardown, 
			(void *)NULL);
	pthread_cleanup_push((void (*)(void *))q_free, 
			(void *)&commsq);
	pthread_cleanup_push((void (*)(void *))q_free, 
			(void *)&hostsq);
	pthread_cleanup_push((void (*)(void *))hash_free, 
			(void *)&hostsh);
	pthread_cleanup_push((void (*)(void *))hash_free, 
			(void *)&cbh);
	pthread_cleanup_push(thread_exec_cleanup_comms, 
			(void *)comms);
	pthread_cleanup_push(thread_exec_cleanup_hosts, 
			(void *)hosts);

	slots = 0;
	pend = 0;
	width = 0;

	while (width == 0) {
		sleep(1);
		thread_exec_fill(hostsh, hostsq, hosts, &width, &slots);
	}

thread_exec_head:

	/* FIXME: make configurable. */
	if (width < opts.ptrigger) for (;;) {
		thread_exec_fill(hostsh, hostsq, hosts, 
			&width, &slots);
		if (0 < slots)
			break;
		sleep(1);
	}

	time_s(cur);

	for ( ; pend < opts.ppar; ) {
		if (NULL == (h = (struct host_t *)q_head(&hostsq)))
			break;
		if (0 != h->last && cur < h->last + (time_t)opts.pwait)
			break;
		h = q_rem(&hostsq);
		assert(NULL != h);
		while (opts.pburst > h->pending && 0 < h->queue_sz) {
			p = q_rem(&h->queue);
			assert(NULL != p);
			--(h->queue_sz);
			++(h->pending);

			dbg("scheduling (pend: %zd, slots: %zd, "
					"width: %zd): %s", pend, 
					slots, width,  uri_topstring_r
					(&p->record.uri, addr));

			p->record.size = 0;
			p->record.expires = 0;
			p->record.type = NLPCDBR_TYPE_NONE;
			memset(p->record.charset, 0, NLPCR_CHARSETLEN);
			memset(p->record.lang, 0, NLPCR_LANGLEN);

			c = q_rem(&commsq);
			assert(NULL != c);
			comm_open(c, opts.cache, p);
			if (opts.ppar <= ++pend)
				break;
		}
	}

	comm_wait();

	pthread_testcancel();

	cookie = comm_perform();
	time_s(cur);

	while (NULL != (handle = comm_report_next(&cookie, &type))) {
		hash_get(&cbh, &handle, sizeof(void *), (void **)&c);
		assert(NULL != c);
		auth = c->pair->record.uri.auth;

		hash_get(&hostsh, auth, NLPCR_URI_AUTHSZ, (void **)&h);
		assert(NULL != h);
		--(h->pending);
		h->last = cur;

		if (0 == h->pending && 0 == h->queue_sz) {
			--width;
			h->valid = 0;
			hash_del(&hostsh, auth, NLPCR_URI_AUTHSZ);
		} else
			q_add(&hostsq, h);

		--slots;
		--pend;

		dbg("unscheduling (pend: %zd, slots: %zd, "
				"width: %zd): %s", pend, slots, 
				width,  uri_topstring_r
				(&c->pair->record.uri, addr));

		q_add(&commsq, c);
		thread_exec_rm(c, type);
	}

	goto thread_exec_head;

	pthread_cleanup_pop(1);
	pthread_cleanup_pop(1);
	pthread_cleanup_pop(1);
	pthread_cleanup_pop(1);
	pthread_cleanup_pop(1);
	pthread_cleanup_pop(1);
	pthread_cleanup_pop(1);

	return(NULL);
}


void *
thread_write(void *arg)
{
	struct nlpcdbrp	*p;
	struct uri	*addr;
	char		buf[NLPCR_URI_MAXSZ + 10],
			reason[32];
	size_t		n, u, d, f;
	int		wflags, rflags;

	assert(NULL == arg);

	wflags = SQUEUE_WAIT | SQUEUE_NOTIFY;
	rflags = SQUEUE_NOTIFY;
	u = f = n = d = 0;

	for (;;) {
		p = squeue_rem(opts.wqueue, rflags);
		if (NULL == p) {
			if (u > 0) 
				signal_next(NLPCDBR_UNREACH);
			if (f > 0)
				signal_next(NLPCDBR_FRESH);
			if (n > 0)
				signal_next(NLPCDBR_NEW);
			if (d > 0)
				signal_next(NLPCDBR_DEAD);
			u = d = f = n = 0;
			ddbg("%s: signalled allies", __func__);
			p = squeue_rem(opts.wqueue, wflags);
		}
		assert(NULL != p);
		addr = &p->record.uri;
		uri_topstring_r(addr, buf);
		nlpcdbr_code2str_r(reason, p->record.code);

		if (NLPCDBR_NEW == p->record.state && ! dict_has
				(&opts.scheme, addr->scheme, 
				 strlen(addr->scheme))) {
			dbg("reject: non-matching scheme: %s", buf);
			goto thread_write_tail;
		}
		if (NLPCDBR_NEW == p->record.state &&  ! dict_has
				(&opts.auth, addr->auth,
				 strlen(addr->auth))) {
			dbg("reject: non-matching auth: %s", buf);
			goto thread_write_tail;
		}

		if (NLPCDBR_NEW == p->record.state) {
			if ( ! db_add(p)) {
				ddbg("reject: duplicate: %s", buf);
				goto thread_write_tail;
			}
		} else if ( ! db_put(p))
			diex("%s: fail: db_put: %s", __func__, buf);

		switch (p->record.state) {
		case (NLPCDBR_UNREACH):
			u++;
			dbg("accept: unreach (%s): %s", reason, buf);
			break;
		case (NLPCDBR_FRESH):
			f++;
			dbg("accept: mod-fresh: %s", buf);
			break;
		case (NLPCDBR_NEW):
			n++;
			dbg("accept: new: %s", buf);
			break;
		case (NLPCDBR_DEAD):
			d++;
			dbg("accept: dead (%s): %s", reason, buf);
			break;
		default:
			abort();
		}

thread_write_tail:	
		free(p);
	}

	return(NULL);
}


static int
scan(void)
{
	struct nlpcdbrp	*p;
	struct db_query	query;
	void		*cookie;
	int		flags, rc;
	static uint64_t	last;

	memset(&query, 0, sizeof(struct db_query));

	query.field = DB_FIELD_STATE;
	query.state = NLPCDBR_NEW | NLPCDBR_DIRTY;
	query.skip = last;

	db_window_fill(&query);
	db_query_begin(&cookie);

	for (rc = 1, p = NULL, flags = 0; ; ) {
		if (SIGNAL_EXIT & (flags |= signal_test())) 
			break;
		calloc_s(p, 1, sizeof(struct nlpcdbrp));
		if (1 != (rc = db_query_next(&cookie, &query, p)))
			break;
		if ( ! squeue_add(opts.rqueue, p, SQUEUE_NOTIFY))
			break;
		p = NULL;
	}

	db_query_close(&cookie);
	last = (0 == rc ? 0 : query.pos);
	if (NULL != p)
		free(p);

	return(flags);
}


int
main(int argc, char *argv[])
{
	int 		c, t;
	char 		*dir;

	extern char	*__progname, *optarg;
	extern int	optind;
	pthread_t	writer, execer;

	static struct option longopts[] = {
		{ "head-lang", 		required_argument, NULL, 1 },
		{ "head-charset", 	required_argument, NULL, 2 },
		{ "filter-charset", 	required_argument, NULL, 7 },
		{ "filter-lang", 	required_argument, NULL, 8 },
		{ "filter-scheme", 	required_argument, NULL, 9 },
		{ "filter-auth", 	required_argument, NULL, 10 },
		{ "filter-ascend", 	no_argument, NULL, 11 },
		{ "fetch-burst",	required_argument, NULL, 13 },
		{ "fetch-depth",	required_argument, NULL, 14 },
		{ "fetch-width",	required_argument, NULL, 15 },
		{ "fetch-wait",		required_argument, NULL, 16 },
		{ "fetch-parallel",	required_argument, NULL, 17 },
		{ NULL, 0, NULL, 0 }
	};

	memset(&opts, 0, sizeof(struct opt_t));

	opts.cache = NLPC_DEFAULT_FILES;
	opts.pburst = FETCH_QUEUE_BURST_SZ;
	opts.pdepth = FETCH_QUEUE_BURST_MAX;
	opts.pwait = FETCH_QUEUE_BURST_WAIT;
	opts.pwidth = FETCH_QUEUE_SZ;
	opts.ppar = FETCH_DEF_COMM;

	dir = NLPC_DEFAULT_DIR;

	while (-1 != (c = getopt_long(argc, argv, "b:hdvf:p:", 
					longopts, NULL))) 
		switch (c) {
		case (1):
			comm_hdr_lang(optarg);
			break;
		case (2):
			comm_hdr_charset(optarg);
			break;
		case (7):
			dict_free(&opts.enc);
			dict_fill(&opts.enc, optarg);
			break;
		case (8):
			dict_free(&opts.lang);
			dict_fill(&opts.lang, optarg);
			break;
		case (9):
			dict_free(&opts.scheme);
			dict_fill(&opts.scheme, optarg);
			break;
		case (10):
			dict_free(&opts.auth);
			dict_fill(&opts.auth, optarg);
			break;
		case (11):
			opts.ascend = 1;
			break;
		case (12):
			if ( ! parse_int(optarg, &t, INT_MAX, 1))
				diex("invalid value: --fetch-trigger");
			opts.ptrigger = (size_t)t;
			break;
		case (13):
			if ( ! parse_int(optarg, &t, INT_MAX, 1))
				diex("invalid value: --fetch-burst");
			opts.pburst = (size_t)t;
			break;
		case (14):
			if ( ! parse_int(optarg, &t, INT_MAX, 1))
				diex("invalid value: --fetch-depth");
			opts.pdepth = (size_t)t;
			break;
		case (15):
			if ( ! parse_int(optarg, &t, INT_MAX, 1))
				diex("invalid value: --fetch-width");
			opts.pwidth = (size_t)t;
			break;
		case (16):
			if ( ! parse_int(optarg, &t, INT_MAX, 0))
				diex("invalid value: --fetch-wait");
			opts.pwait = (size_t)t;
			break;
		case (17):
			if ( ! parse_int(optarg, &t, INT_MAX, 0))
				diex("invalid value: --fetch-parallel");
			opts.pwait = (size_t)t;
			break;
		case ('d'):
			dbg_lvl++;
			break;
		case ('f'):
			opts.cache = optarg;
			break;
		case ('p'):
			dir = optarg;
			break;
		case ('v'):
			fprintf(stdout, "%s version %s\n", 
					__progname, VERSION);
			return(EXIT_SUCCESS);
		case ('h'):
			/*FALLTHROUGH*/
		default:
			fprintf(stderr, "usage: %s [-dhv] [-f path] "
					"[-p path] [LONG_OPTS]\n", 
					__progname);
			return(EXIT_FAILURE);
		}
	argc -= optind;
	argv += optind;

	assert(opts.pwidth > 0);
	assert(opts.pdepth > 0);
	assert(opts.ppar > 0);
	assert(opts.pburst > 0);
	assert(opts.pwait > 0);

	if (0 == opts.ptrigger)
		opts.ptrigger = opts.pwidth / 4;
	if (opts.ptrigger > (opts.pwidth * opts.pdepth))
		diex("value overruns queue domain: --fetch-trigger");

	opts.wqueue = squeue_new(opts.pdepth * opts.pwidth);
	opts.rqueue = squeue_new(opts.pdepth * opts.pwidth);

	signal_init(dir, NLPCDBR_NEW | NLPCDBR_DIRTY);
	db_init(dir, 0);

	if (0 != pthread_create(&writer, NULL, thread_write, NULL))
		die("%s: fail: pthread_create", __func__);
	if (0 != pthread_create(&execer, NULL, thread_exec, NULL))
		die("%s: fail: pthread_create", __func__);

	comm_hdr_agent(VERSION);

	for (;;) {
		if (SIGNAL_EXIT & scan())
			break;
		if (SIGNAL_EXIT & signal_wait(0))
			break;
	}

	if (0 != pthread_cancel(writer))
		die("%s: fail: pthread_cancel", __func__);
	if (0 != pthread_cancel(execer))
		die("%s: fail: pthread_cancel", __func__);

	if (0 != pthread_join(writer, NULL))
		die("%s: fail: pthread_join", __func__);
	if (0 != pthread_join(execer, NULL))
		die("%s: fail: pthread_join", __func__);

	db_teardown();
	signal_teardown();

	dict_free(&opts.enc);
	dict_free(&opts.lang);
	dict_free(&opts.auth);
	dict_free(&opts.scheme);

	squeue_free(opts.wqueue);
	squeue_free(opts.rqueue);

	return(EXIT_SUCCESS);
}

