/* $Id: list.c,v 1.63 2007-05-25 10:31:03 kristaps Exp $ */
/*
Copyright (c) 2007 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/param.h>
#include <sys/ioctl.h>

#include <assert.h>
#include <err.h>
#include <getopt.h>
#include <inttypes.h>
#include <limits.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "nlpcrawl.h"


#define LIST_DEF_WIDTH	132
#define LIST_MIN_WIDTH	80


static struct opt_t {
	int		quiet;		/* Quiet (no record-print).  */
	int		longform;	/* Show details (multiline). */
	unsigned short	cols;           /* Number of output columns. */
	int		tree;		/* If report hash-tree.      */
	int		dupes;		/* If report content-dupes.  */
} opts;


static struct sstats_t {
	uint64_t	new;		/* Number of new records.    */
	uint64_t	fresh;		/* Number of fresh records.  */
	uint64_t	dirty;		/* Number of dirty records.  */
	uint64_t	unreach;	/* Number of unreach records.*/
	uint64_t	dead;		/* Number of dead records.   */
	uint64_t	scan;		/* Number of scan records.   */
	uint64_t	vrfy;		/* Number of vrfy records.   */
	uint64_t	total;		/* Total number of records.  */
	uint64_t	sz_total;	/* Total size registered.    */
	uint64_t	sz_fresh;	/* Size of fresh pages.      */
	uint64_t	sz_dirty;	/* Size of dirty pages.      */
	uint64_t	sz_unreach;	/* Size of unreach pages.    */
	uint64_t	sz_dead;	/* Size of dead pages.       */
	uint64_t	sz_scan;	/* Size of scanned pages.    */
	uint64_t	sz_vrfy;	/* Size of verified pages.   */
	uint64_t	dupes;		/* Number of dupe-contents.  */
	uint64_t	dupes_sz;	/* Number of dupe-checks.    */
	void		*dict_dupes;	/* Dict of content-hashes.   */
} stats;



static void
update_sstats(struct nlpcdbrp *pair)
{
	struct nlpcdbr	*p;

	assert(NULL != pair);
	p = &pair->record;

	stats.sz_total += p->size;
	stats.total++;

	switch (p->state) {
	case (NLPCDBR_NEW):
		stats.new++;
		break;
	case (NLPCDBR_FRESH):
		stats.sz_fresh += p->size;
		stats.fresh++;
		break;
	case (NLPCDBR_DIRTY):
		stats.sz_dirty += p->size;
		stats.dirty++;
		break;
	case (NLPCDBR_UNREACH):
		stats.sz_unreach += p->size;
		stats.unreach++;
		break;
	case (NLPCDBR_DEAD):
		stats.sz_dead += p->size;
		stats.dead++;
		break;
	case (NLPCDBR_SCAN):
		stats.sz_scan += p->size;
		stats.scan++;
		break;
	case (NLPCDBR_VRFY):
		stats.sz_vrfy += p->size;
		stats.vrfy++;
		break;
	default:
		abort();
	}

	if (opts.dupes && 0 != p->data[0]) {
		stats.dupes_sz++;
		if (dict_has(&stats.dict_dupes, p->data, NLPCR_HASHLEN))
			stats.dupes++;
		else
			dict_put(&stats.dict_dupes, p->data, 
					NLPCR_HASHLEN);
	}
}


static void
show_stats(void)
{
	printf("%9s: %-8" PRIu64, 
		nlpcdbr_state2str(NLPCDBR_NEW), stats.new);
	printf("%9s: %-8" PRIu64, 
		nlpcdbr_state2str(NLPCDBR_FRESH), stats.fresh);
	printf("%9s: %-8" PRIu64, 
		nlpcdbr_state2str(NLPCDBR_DIRTY), stats.dirty);
	printf("%9s: %-8" PRIu64 "\n", 
		nlpcdbr_state2str(NLPCDBR_UNREACH), stats.unreach);

	printf("%9s: %-8" PRIu64,
		nlpcdbr_state2str(NLPCDBR_DEAD), stats.dead);
	printf("%9s: %-8" PRIu64,
		nlpcdbr_state2str(NLPCDBR_SCAN), stats.scan);
	printf("%9s: %-8" PRIu64,
		nlpcdbr_state2str(NLPCDBR_VRFY), stats.vrfy);
	printf("%9s: %-8" PRIu64 "\n",
		"total", stats.total);

	printf("%9s: %-8.2f", "valid", ((stats.scan + stats.vrfy) / 
		(double)(stats.total + 1)) * 100);
	printf("%9s: %-8.2f", "pending", ((stats.new + stats.dirty) /
		(double)(stats.total + 1)) * 100);
	printf("%9s: %-8.2f", "local", ((stats.total - stats.new) /
		(double)(stats.total + 1)) * 100);
	printf("%9s: %-8.2f\n", "dead", ((stats.unreach + stats.dead) /
		(double)(stats.total + 1)) * 100);

	printf("%9s: %-8" PRIu64, "valid",
		(stats.sz_scan + stats.sz_vrfy) / 1000);
	printf("%9s: %-8" PRIu64, "dead",
		(stats.sz_unreach + stats.sz_dead) / 1000);
	printf("%9s: %-8" PRIu64, "new", stats.sz_fresh / 1000);
	printf("%9s: %-8" PRIu64 "\n", "total", stats.sz_total / 1000);

	if ( ! opts.dupes)
		return;

	printf("%9s: %-8.2f\n", "dupes", (stats.dupes /
				(double)(stats.dupes_sz + 1)) * 100);
}


static void
show_long(unsigned long n, const struct nlpcdbrp *pair)
{
	time_t		ftime, atime, rtime, mtime, exp;

	ftime = (time_t)pair->record.ftime;
	atime = (time_t)pair->record.atime;
	mtime = (time_t)pair->record.mtime;
	rtime = (time_t)pair->record.ctime;
	exp = (time_t)pair->record.expires;

	printf("%-8lu%32s (%s)\n", n, nlpcdbr_hash2str(pair->hash), 
			nlpcdbr_state2str(pair->record.state));
	printf("        uri:      %s\n", 
			uri_topstring(&pair->record.uri));
	printf("        data:     %s\n", 0 == pair->record.data[0] ?
				"none" : 
				nlpcdbr_hash2str(pair->record.data));
	printf("        content:  %s, %s\n", 
			nlpcdbr_type2str(pair->record.type), 
			pair->record.charset[0] == 0 ? "none" : 
				pair->record.charset);
	printf("        size:     %" PRIu64 " bytes (%" PRIu64 " KB)\n", 
			pair->record.size, pair->record.size / 1000);
	printf("        lang:     %s\n", 
			pair->record.lang[0] == 0 ? "none" : 
				pair->record.lang);
	printf("        ftime:    %s", 0 == ftime ?
			"none\n" : ctime(&ftime));
	printf("        atime:    %s", 0 == atime ?
			"none\n" : ctime(&atime));
	printf("        ctime:    %s", 0 == rtime ?
			"none\n" : ctime(&rtime));
	printf("        mtime:    %s", 0 == mtime ?
			"none\n" : ctime(&mtime));
	printf("        code:     %s\n", 
			nlpcdbr_code2str(pair->record.code));
	if (0 == exp)
		printf("        expires:  none\n");
	else if (UINT64_MAX == pair->record.expires)
		printf("        expires:  immediate\n");
	else
		printf("        expires:  %s", ctime(&exp));
}


static void
show(unsigned long n, const struct nlpcdbrp *pair)
{
	char 		line[opts.cols], *p;
	unsigned short	mark;

	p = line;
	memset(line, 0, opts.cols);
	snprintf(line, opts.cols - 1, "%-8lu%32s%14s ", n, 
			nlpcdbr_hash2str(pair->hash),
			nlpcdbr_state2str(pair->record.state));
	p = line + strlen(line);
	mark = opts.cols - strlen(line);

	switch (pair->record.state) {
	case (NLPCDBR_NEW):
		break;
	case (NLPCDBR_SCAN):
		/*FALLTHROUGH*/
	case (NLPCDBR_VRFY):
		/*FALLTHROUGH*/
	case (NLPCDBR_FRESH):
		snprintf(p, mark - 1, "(%s; %s) ", 
				nlpcdbr_type2str(pair->record.type),
				pair->record.charset[0] == 0 ? 
				"none" : pair->record.charset);
		break;
	case (NLPCDBR_DIRTY):
		break;
	case (NLPCDBR_UNREACH):
		snprintf(p, mark - 1, "(%s) ", 
				nlpcdbr_code2str(pair->record.code));
		break;
	case (NLPCDBR_DEAD):
		snprintf(p, mark - 1, "(%s) ", 
				nlpcdbr_code2str(pair->record.code));
		break;
	}
	p = line + strlen(line);
	mark = opts.cols - strlen(line);

	snprintf(p, mark - 1, "%s", 
			uri_topstring(&pair->record.uri));

	if (0 == line[opts.cols - 2]) {
		line[opts.cols - 2] = '.';
		line[opts.cols - 3] = '.';
		line[opts.cols - 4] = '.';
	}

	printf("%s\n", line);
}


static void
list_all(struct db_query *query)
{
	unsigned long	i;
	int		rc;
	struct nlpcdbrp	pair;
	void		*cookie;

	db_query_begin(&cookie);

	i = 1;
	while (1 == (rc = db_query_next(&cookie, query, &pair))) {
		update_sstats(&pair);
		if ( ! opts.quiet) {
			opts.longform ? show_long(i++, &pair) : 
				show(i++, &pair);
		}
	}
	assert(-1 != rc);

	db_query_close(&cookie);
}


static void
list(struct db_query *query, int argc, char *argv[])
{
	int		i;
	unsigned long	pos;
	struct nlpcdbrp	pair;
	char		*p;
	void		*cookie;

	assert(NULL != query);

	if (0 == argc) {
		list_all(query);
		return;
	}

	db_query_begin(&cookie);

	for (pos = 0, i = 0; i < argc; i++) {
		if (NULL == (p = nlpcdbr_str2hash(argv[i]))) {
			dbg("reject: invalid hash: %s", argv[i]);
			continue;
		}

		memset(&pair, 0, sizeof(struct nlpcdbrp));
		memcpy(pair.hash, p, NLPCR_HASHLEN);

		if ( ! db_query_lookup(&cookie, query, &pair))
			continue;
		update_sstats(&pair);
		if ( ! opts.quiet) {
			opts.longform ? show_long(pos++, &pair) : 
				show(pos++, &pair);
		}
	}

	db_query_close(&cookie);
}


int
main(int argc, char *argv[])
{
	extern char	*__progname, *optarg;
	extern int 	optind;

	uint64_t	st;
	int 		c;
	char 		*dir, *p;
	struct db_query	query;
	struct winsize	ws;

	static struct option longopts[] = {
		{ "filter-state", 	required_argument, NULL, 1 },
		{ "filter-auth", 	required_argument, NULL, 2 },
		{ "report-tree",	no_argument, NULL, 3 },
		{ "report-dupes",	no_argument, NULL, 4 },
		{ NULL, 0, NULL, 0 }
	};

	memset(&stats, 0, sizeof(struct sstats_t));
	memset(&opts, 0, sizeof(struct opt_t));
	memset(&query, 0, sizeof(struct db_query));

	dict_new(&stats.dict_dupes);

	dir = NLPC_DEFAULT_DIR;
	query.field = DB_FIELD_NONE;
	query.window = -1;

	while (-1 != (c = getopt_long(argc, argv, "whdvlp:q",
					longopts, NULL)))
		switch (c) {
		case (1):
			query.field |= DB_FIELD_STATE;
			query.state = 0;
			while (NULL != (p = strsep(&optarg, ", "))) {
				if (0 == strlen(p))
					continue;
				if (0 == (st = nlpcdbr_str2state(p)))
					diex("invalid state: %s", p);
				query.state |= st;
			}
			break;
		case (2):
			query.field |= DB_FIELD_AUTH;
			if ( ! uri_parse(optarg, 0, &query.addr))
				diex("invalid uri: %s", optarg);
			break;
		case (3):
			opts.tree = 1;
			break;
		case (4):
			opts.dupes = 1;
			break;
		case ('l'):
			opts.longform = 1;
			break;
		case ('d'):
			dbg_lvl++;
			break;
		case ('p'):
			dir = optarg;
			break;
		case ('q'):
			opts.quiet = 1;
			break;
		case ('v'):
			fprintf(stdout, "%s version %s\n", 
					__progname, VERSION);
			exit(EXIT_SUCCESS);
		case ('w'):
			opts.cols = (0 == opts.cols) ? 
				LIST_DEF_WIDTH : 4096;
			break;
		case ('h'):
			/*FALLTHROUGH*/
		default:
			fprintf(stderr, "usage: %s [-dhlqv] [-p path] "
					"[LONG_OPTS]\n", __progname);
			exit(EXIT_FAILURE);
		}
	argc -= optind;
	argv += optind;

	if (0 == opts.cols) {
		if (ioctl(1, TIOCGWINSZ, (void *)&ws) != 0) {
			dbg("warn: failed terminal width");
			opts.cols = LIST_DEF_WIDTH;
		} else
			opts.cols = ws.ws_col;
	}

	signal_init(dir, 0);
	db_init(dir, NLPCDB_RDONLY);

	list(&query, argc, argv);
	show_stats();

	dict_free(&stats.dict_dupes);

	db_teardown();
	signal_teardown();
	return(EXIT_SUCCESS);
}

