/* $Id: mt.c,v 1.6 2007-05-24 16:08:55 kristaps Exp $ */
/*
Copyright (c) 2007 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/param.h>

#include <assert.h>
#include <err.h>
#include <errno.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "nlpcrawl.h"


struct squeue
{
	void		*queue;
	pthread_mutex_t	mux;
	pthread_cond_t	cond;
	size_t		max;
	size_t		size;
};


static void
mt_cond_bcast(pthread_cond_t *p)
{
	assert(NULL != p);
	if (0 != pthread_cond_broadcast(p))
		die("%s: fail: pthread_cond_bcast", __func__);
}


static void
mt_cond_wait(pthread_mutex_t *t, pthread_cond_t *c)
{
	assert(NULL != t);
	assert(NULL != c);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
	if (0 != pthread_cond_wait(c, t))
		die("%s: fail: pthread_cond_wait", __func__);
	pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);
}


#define mt_mux_lock(p) \
	pthread_cleanup_push((void (*)(void *))pthread_mutex_unlock, (void *)(p)); \
	if (0 != pthread_mutex_lock((p))) \
		die("%s: fail: pthread_mutex_lock", __func__); \

#define mt_mux_unlock()  \
	pthread_cleanup_pop(1)


void *
squeue_new(size_t max)
{
	struct squeue	*p;

	if (NULL == (p = calloc(1, sizeof(struct squeue))))
		die("%s: fail: allocate", __func__);

	q_new(&p->queue);
	pthread_mutex_init(&p->mux, NULL);
	pthread_cond_init(&p->cond, NULL);

	p->max = max;
	p->size = 0;

	return(p);
}


void
squeue_free(void *cookie)
{
	struct squeue	*p;

	if (NULL == cookie)
		return;

	p = (struct squeue *)cookie;

	q_free(&p->queue);
	pthread_cond_destroy(&p->cond);
	pthread_mutex_destroy(&p->mux);

	free(p);
}


void
squeue_wait(void *cookie)
{
	struct squeue	*p;
	assert(NULL != cookie);
	p = (struct squeue *)cookie;

	mt_mux_lock(&p->mux);
	mt_cond_wait(&p->mux, &p->cond);
	mt_mux_unlock();
}


void
squeue_notify(void *cookie)
{
	struct squeue	*p;
	assert(NULL != cookie);
	p = (struct squeue *)cookie;
	mt_cond_bcast(&p->cond);
}


int
squeue_add(void *cookie, void *data, uint64_t flags)
{
	struct squeue	*p;
	int		rc;

	assert(NULL != cookie);
	assert(NULL != data);

	p = (struct squeue *)cookie;

	mt_mux_lock(&p->mux);

squeue_add_head:

	rc = 1;

	if (0 == p->max || p->max > p->size) {
		(p->size)++;
		q_add(&p->queue, data);
		if (flags & SQUEUE_NOTIFY)
			mt_cond_bcast(&p->cond);
	} else if (flags & SQUEUE_WAIT) {
		mt_cond_wait(&p->mux, &p->cond);
		goto squeue_add_head;
	} else {
		rc = 0;
	}

	mt_mux_unlock();
	return(rc);
}


void *
squeue_rem(void *cookie, uint64_t flags)
{
	struct squeue	*p;
	void		*res;
	size_t		ssz;

	assert(NULL != cookie);

	p = (struct squeue *)cookie;

	mt_mux_lock(&p->mux);

squeue_rem_head:

	res = NULL;

	if ( ! q_empty(&p->queue)) {
		res = q_rem(&p->queue);
		ssz = (p->size)--;
		if (flags & SQUEUE_NOTIFY && 
				0 != p->max && p->max == ssz)
			mt_cond_bcast(&p->cond);
	} else if (flags & SQUEUE_WAIT) {
		mt_cond_wait(&p->mux, &p->cond);
		goto squeue_rem_head;
	}

	mt_mux_unlock();
	return(res);
}


