/* $Id: nlpc.c,v 1.30 2007-05-24 16:08:55 kristaps Exp $ */
/*
Copyright (c) 2007 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/param.h>
#include <arpa/inet.h>

#include <assert.h>
#include <err.h>
#include <errno.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "nlpcrawl.h"

/* 
 * Default debugging level is set to display no messages.
 */
int			dbg_lvl = 0;

static pthread_mutex_t	mux = PTHREAD_MUTEX_INITIALIZER;


void
nlpcr_new(const struct uri *addr, struct nlpcdbrp *res)
{
	assert(NULL != res);
	assert(NULL != addr);

	md5_hash(uri_tostring(addr), res->hash);

	memset(&res->record, 0, sizeof(struct nlpcdbr));

	memcpy(&res->record.uri, addr, sizeof(struct uri));

	res->record.state = NLPCDBR_NEW;
	res->record.code = NLPCDBR_OK;
	res->record.type = NLPCDBR_TYPE_NONE;
	res->record.ftime = 0;
	res->record.atime = res->record.mtime =
		res->record.ctime = time_uint64();
}


void
vdbg(int level, const char *fmt, va_list ap)
{
	extern char 	*__progname;

	if (level >= dbg_lvl)
		return;

	if (0 != pthread_mutex_lock(&mux))
		die("%s: fail: pthread_mutex_lock", __func__);

	fprintf(stdout, "%s: debug%d: ", __progname, level);
	vprintf(fmt, ap);
	fprintf(stdout, "\n");
	fflush(stdout);

	if (0 != pthread_mutex_unlock(&mux))
		die("%s: fail: pthread_mutex_unlock", __func__);
}


void
dbg(const char *fmt, ...) 
{
	va_list 	ap;
	va_start(ap, fmt);
	vdbg(0, fmt, ap);
	va_end(ap);
}


void
dddbg(const char *fmt, ...) 
{
	va_list 	ap;
	va_start(ap, fmt);
	vdbg(2, fmt, ap);
	va_end(ap);
}


void
ddbg(const char *fmt, ...) 
{
	va_list 	ap;
	va_start(ap, fmt);
	vdbg(1, fmt, ap);
	va_end(ap);
}


uint64_t 
nlpcdbr_str2state(const char *state)
{
	assert(NULL != state);
	if (0 == strcasecmp("new", state))
		return(NLPCDBR_NEW);
	if (0 == strcasecmp("fresh", state))
		return(NLPCDBR_FRESH);
	if (0 == strcasecmp("dirty", state))
		return(NLPCDBR_DIRTY);
	if (0 == strcasecmp("temp", state))
		return(NLPCDBR_UNREACH);
	if (0 == strcasecmp("scan", state))
		return(NLPCDBR_SCAN);
	if (0 == strcasecmp("dead", state))
		return(NLPCDBR_DEAD);
	if (0 == strcasecmp("vrfy", state))
		return(NLPCDBR_VRFY);
	return(0);
}


char *
nlpcdbr_state2str(uint64_t state)
{
	static char str[32];

	return(nlpcdbr_state2str_r(str, state));
}


char *
nlpcdbr_state2str_r(char *str, uint64_t state)
{
	assert(NULL != str);
	memset(str, 0, 32);

	switch (state) {
	case (NLPCDBR_NEW):
		strncpy(str, "new", 32);
		break;
	case (NLPCDBR_FRESH):
		strncpy(str, "fresh", 32);
		break;
	case (NLPCDBR_DIRTY):
		strncpy(str, "dirty", 32);
		break;
	case (NLPCDBR_UNREACH):
		strncpy(str, "temp", 32);
		break;
	case (NLPCDBR_SCAN):
		strncpy(str, "scan", 32);
		break;
	case (NLPCDBR_DEAD):
		strncpy(str, "dead", 32);
		break;
	case (NLPCDBR_VRFY):
		strncpy(str, "vrfy", 32);
		break;
	default:
		strncpy(str, "unknown", 32);
		break;
	}
	return(str);
}


char *
nlpcdbr_type2str(uint64_t type)
{
	static char str[32];

	return(nlpcdbr_type2str_r(str, type));
}


char *
nlpcdbr_type2str_r(char *str, uint64_t type)
{
	assert(NULL != str);
	memset(str, 0, 32);

	switch (type) {
	case (NLPCDBR_TYPE_NONE):
		strncpy(str, "none", 32);
		break;
	case (NLPCDBR_TYPE_TXT_PLAIN):
		strncpy(str, "txt/plain", 32);
		break;
	case (NLPCDBR_TYPE_TXT_XML):
		strncpy(str, "txt/xml", 32);
		break;
	case (NLPCDBR_TYPE_APP_XML):
		strncpy(str, "app/xml", 32);
		break;
	case (NLPCDBR_TYPE_APP_XHTML):
		strncpy(str, "app/xhtml", 32);
		break;
	case (NLPCDBR_TYPE_TXT_HTML):
		strncpy(str, "txt/html", 32);
		break;
	default:
		strncpy(str, "unknown", 32);
		break;
	}
	return(str);
}


char *
nlpcdbr_code2str(uint64_t code)
{
	static char str[32];
	return(nlpcdbr_code2str_r(str, code));
}


char *
nlpcdbr_code2str_r(char *str, uint64_t code)
{
	assert(NULL != str);
	memset(str, 0, 32);

	switch (code) {
	case (NLPCDBR_OK):
		strncpy(str, "no error", 32);
		break;
	case (NLPCDBR_GENERAL):
		strncpy(str, "general error", 32);
		break;
	case (NLPCDBR_RESOLV):
		strncpy(str, "name resolution failed", 32);
		break;
	case (NLPCDBR_CONNECT):
		strncpy(str, "server connect failed", 32);
		break;
	case (NLPCDBR_TIMEO):
		strncpy(str, "connection timeout exceeded", 32);
		break;
	case (NLPCDBR_SERVER):
		strncpy(str, "server-side error", 32);
		break;
	case (NLPCDBR_SYNTAX):
		strncpy(str, "malformed content syntax", 32);
		break;
	case (NLPCDBR_URL):
		strncpy(str, "malformed uri syntax", 32);
		break;
	case (NLPCDBR_MOVED):
		strncpy(str, "content moved", 32);
		break;
	case (NLPCDBR_CTYPE):
		strncpy(str, "invalid content type", 32);
		break;
	case (NLPCDBR_CHARSET):
		strncpy(str, "invalid charset", 32);
		break;
	case (NLPCDBR_LANG):
		strncpy(str, "invalid language", 32);
		break;
	case (NLPCDBR_RTD):
		strncpy(str, "revivals exceeded", 32);
		break;
	case (NLPCDBR_LIMIT):
		strncpy(str, "download limit exceeded", 32);
		break;
	default:
		strncpy(str, "unknown code", 32);
		break;
	}
	return(str);
}


char *
nlpcdbr_str2hash(const char *str)
{
	static char	hash[NLPCR_HASHLEN];
	return(nlpcdbr_str2hash_r(hash, str));
}


char *
nlpcdbr_str2hash_r(char *hash, const char *str)
{
	size_t		sz;
	char		hex[5];
	int		j, hexv;

	assert(NULL != hash);
	assert(NULL != str);

	if (NLPCR_HASHLEN * 2 != (sz = strlen(str))) {
		dddbg("%s: invalid hash length: %zd, %s", 
				__func__, sz, str);
		return(NULL);
	}

	for (j = 0; j < NLPCR_HASHLEN; j++) {
		hex[0] = '0';
		hex[1] = 'x';
		hex[2] = str[j * 2];
		hex[3] = str[(j * 2) + 1];
		hex[4] = 0;
		if (parse_hex(hex, &hexv, UCHAR_MAX, 0)) {
			hash[j] = (unsigned char)hexv;
			continue;
		}
		dddbg("%s: invalid hash sequence: %s, %s",
				__func__, hex, str);
		return(NULL);
	}

	return(hash);
}


char *
nlpcdbr_hash2str(const unsigned char hash[NLPCR_HASHLEN])
{
	static char	str[(NLPCR_HASHLEN * 2) + 1];
	return(nlpcdbr_hash2str_r(str, hash));
}


char *
nlpcdbr_hash2str_r(char *str, const unsigned char hash[NLPCR_HASHLEN])
{
	int		i;

	assert(NULL != str);

	for (i = 0; i < NLPCR_HASHLEN; i++) 
		snprintf(&str[i * 2], 3, "%.2x", hash[i]);
	str[NLPCR_HASHLEN * 2] = 0;

	return(str);
}


char *
nlpcdbr_hash2path(const char *prefix, 
		const unsigned char hash[NLPCR_HASHLEN])
{
	static char	path[MAXPATHLEN];
	return(nlpcdbr_hash2path_r(path, prefix, hash));
}


char *
nlpcdbr_hash2path_r(char *path, const char *prefix, 
		const unsigned char hash[NLPCR_HASHLEN])
{
	char		*p;
	int		i;

	assert(NULL != prefix);
	assert(NULL != path);

	memset(path, 0, MAXPATHLEN);

	strncat(path, prefix, MAXPATHLEN);
	if (0 != path[MAXPATHLEN - 1])
		diex("%s: fail: path prefix too long: %s", 
				__func__, prefix);

	if (MAXPATHLEN - 1 < strlen(path) + (3 * NLPCR_HASHLEN))
		diex("%s: fail: filename too long: %s", 
				__func__, prefix);

	p = path + strlen(path);

	for (i = 0; i < NLPCR_HASHLEN; i++) {
		snprintf(p, 4, "/%.2x", hash[i]);
		p += 3;
	}

	return(path);
}
