/* $Id: scan.c,v 1.95 2007-05-25 10:31:03 kristaps Exp $ */
/*
Copyright (c) 2007 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/param.h>

#include <assert.h>
#include <err.h>
#include <ctype.h>
#include <getopt.h>
#include <inttypes.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <libxml/HTMLparser.h>
#include <libxml/HTMLtree.h>

#include "nlpcrawl.h"


/* Tag structures relevant to parsing sequence. */
#define XML_HREF	((xmlChar *)"href")
#define XML_ATAG	((xmlChar *)"a")
#define XML_LINKTAG	((xmlChar *)"link")
#define XML_BASETAG	((xmlChar *)"base")


static struct opt_t {
	void		*scheme;	/* Htable of schemes.        */
	void		*auth;		/* Htable of authorities.    */
	void		*q_new;		/* Write-queue for fresh.    */
	void		*q_mod;		/* Write-queue for mods.     */
	char		*cache;		/* Cache-file directory.     */
	int		ascend;		/* Perform path-ascendence.  */
	int		ascend_scan;	/* Path ascend only vrfy.    */
} opts;


/*
 * Enqueue a new entry.  If count is non-NULL, increment it with each
 * new addition.
 */
static void
proc_html_enqueue(struct uri *addr, uint64_t *count)
{
	struct nlpcdbrp	*n;

	assert(NULL != addr);

	calloc_s(n, 1, sizeof(struct nlpcdbrp));
	nlpcr_new(addr, n);
	q_add(&opts.q_new, n);
	if (NULL != count)
		(*count)++;
}


/*
 * Register a parsed link (if applicable) with the queue of new links.
 */
static void
proc_html_link(const struct uri *baseuri, uint64_t *count, 
		const char *link)
{
	struct uri	p, addr;

	assert(NULL != link);

	if (NULL != baseuri) {
		if ( ! uri_parse(link, 0, &p)) {
			dbg("reject: malformed uri: %s", link);
			return;
		} else if ( ! uri_ref_resolve(baseuri, &p, &addr)) {
			dbg("reject: failed reference resolution: %s",
					uri_topstring(&p));
			return;
		}
	} else if ( ! uri_parse(link, 0, &addr)) {
		dbg("reject: malformed uri: %s", link);
		return;
	} else if ( ! uri_is_abs(&addr)) {
		dbg("reject: uri not absolute: %s", 
				uri_topstring(&addr));
		return;
	}

	if ( ! dict_has(&opts.scheme, addr.scheme, 
				strlen(addr.scheme))) {
		dbg("reject: non-matching scheme: %s", 
				uri_topstring(&addr));
		return;
	}
	if ( ! dict_has(&opts.auth, addr.auth,
				strlen(addr.auth))) {
		dbg("reject: non-matching authority: %s", 
				uri_topstring(&addr));
		return;
	}

	memset(addr.fragment, 0, NLPCR_URI_FRAGSZ);
	if (opts.ascend) {
		while (uri_ascend(&addr))
			proc_html_enqueue(&addr, count);
	} else
		proc_html_enqueue(&addr, count);
}


static void
proc_baseuri(xmlChar *str, struct uri *baseuri)
{
	struct uri 	p;

	assert(NULL != baseuri);
	/*assert(NULL != str);*/

	if (NULL == str)
		return;
	if (uri_parse((char *)str, 0, &p)) {
		if ( ! uri_is_abs(&p))
			ddbg("warn: uri not absolute: %s",
					uri_topstring(&p));
		else	memcpy(baseuri, &p, sizeof(struct uri));
	} else
		ddbg("warn: uri failed parse: %s", (char *)str);

	xmlFree(str);
}


static void
proc_html_node(struct uri *baseuri, uint64_t *links, xmlNodePtr node)
{
	xmlNodePtr 	c;
	xmlChar 	*ref;
	const xmlChar	*p;

	assert(NULL != baseuri);
	assert(NULL != links);
	/*assert(NULL != node);*/

	for (c = node; NULL != c; c = c->next) {
		proc_html_node(baseuri, links, c->children);
		if (c->type != XML_ELEMENT_NODE)
			continue;
		p = c->name;

		if (xmlStrcasecmp(XML_BASETAG, p) == 0)
			proc_baseuri(xmlGetProp(c, XML_HREF), baseuri);
		if (xmlStrcasecmp(XML_ATAG, p) != 0 &&
				xmlStrcasecmp(XML_LINKTAG, p) != 0)
			continue;

		if (NULL == (ref = xmlGetProp(c, XML_HREF))) 
			continue;
		proc_html_link(baseuri, links, (const char *)ref);
		xmlFree(ref);
	}
}


/*
 * Return 0 on skip, 1 on success, -1 on failure.
 */
static int
proc_html(struct nlpcdbrp *pair, uint64_t *links)
{
	struct uri	baseuri;
	char		*path;
	struct nlpcdbr	*rec;
	htmlDocPtr 	p;
	xmlNodePtr 	n;

	assert(NULL != pair);
	assert(NULL != links);

	rec = &pair->record;

	assert(0 != rec->charset[0]);
	path = nlpcdbr_hash2path(opts.cache, pair->hash);

	if (NULL == (p = htmlParseFile(path, rec->charset))) {
		rec->code = NLPCDBR_SYNTAX;
		dbg("reject: failed parse: %s", 
				uri_topstring(&rec->uri));
		return(-1);
	}

	memcpy(&baseuri, &rec->uri, sizeof(struct uri));

	if (NULL == (n = xmlDocGetRootElement(p))) {
		rec->code = NLPCDBR_SYNTAX;
		dbg("reject: failed acquire root: %s", 
				uri_topstring(&rec->uri));
		xmlFreeDoc(p);
		return(-1);
	}

	proc_html_node(&baseuri, links, n);
	xmlFreeDoc(p);

	dbg("accept: scan: %s", uri_topstring(&rec->uri));

	return(1);
}


/* 
 * Return 0 on pass, 1 on success (SCAN), -1 on fail (DEAD).
 */
static int
proc(struct nlpcdbrp *pair)
{
	int		rc;
	uint64_t	links;

	assert(NULL != pair);

	links = 0;
	switch (pair->record.type) {
	case (NLPCDBR_TYPE_TXT_HTML):
		if (0 == (rc = proc_html(pair, &links)))
			return(0);
		break;
	case (NLPCDBR_TYPE_TXT_PLAIN):
		rc = 1;
		break;
	default:
		ddbg("warn: unsupported type: %s, %s", 
				nlpcdbr_type2str(pair->record.type), 
				uri_topstring(&pair->record.uri));
		return(0);
	}

	return(rc);
}


static int
scan_mod(void)
{
	int		scan, unreach, flags;
	struct uri	*addr;
	struct nlpcdbrp	*n;

	if (q_empty(&opts.q_mod))
		return(0);

	for (scan = 0, unreach = 0, flags = 0; ; ) {
		n = (struct nlpcdbrp *)q_rem(&opts.q_mod);
		if (NULL == n)
			break;
		addr = &n->record.uri;

		if (SIGNAL_EXIT & (flags |= signal_test()))
			break;

		if ( ! db_put(n))
			diex("fail-modify: %s", uri_topstring(addr));

		switch (n->record.state) {
		case (NLPCDBR_SCAN):
			scan++;
			break;
		case (NLPCDBR_UNREACH):
			unreach++;
			break;
		default:
			abort();
		}

		free(n);
	}

	if (unreach > 0)
		signal_next(NLPCDBR_UNREACH);
	if (scan > 0)
		signal_next(NLPCDBR_SCAN);

	return(flags);
}


static int
scan_new(void)
{
	struct nlpcdbrp	*n;
	struct uri	*addr;
	int		i, flags;

	if (q_empty(&opts.q_new))
		return(0);

	for (flags = 0, i = 0; ; ) {
		n = (struct nlpcdbrp *)q_rem(&opts.q_new);
		if (NULL == n)
			break;
		addr = &n->record.uri;

		if (SIGNAL_EXIT & (flags |= signal_test()))
			break;

		if (db_add(n)) {
			dbg("accept: new: %s", uri_topstring(addr));
			i++;
		} else
			dbg("reject: dupe: %s", uri_topstring(addr));

		free(n);
	}
	if (i > 0)
		signal_next(NLPCDBR_NEW);

	return(flags);
}


static int
scan(void)
{
	void		*cookie;
	struct db_query	query;
	int		rc, flags;
	struct nlpcdbrp pair, *n;

	memset(&query, 0, sizeof(struct db_query));

	query.field = DB_FIELD_STATE;
	query.state = NLPCDBR_VRFY;

	db_window_fill(&query);
	db_query_begin(&cookie);

	for (flags = 0; ; ) {
		if (SIGNAL_EXIT & (flags |= signal_test()))
			break;
		if (1 != db_query_next(&cookie, &query, &pair))
			break;
		if (0 == (rc = proc(&pair)))
			continue;

		pair.record.atime = time_uint64();
		pair.record.mtime = time_uint64();
		pair.record.state = -1 == rc ? 
			NLPCDBR_UNREACH : NLPCDBR_SCAN;

		if (NLPCDBR_SCAN == pair.record.state)
			pair.record.revived = 0;

		calloc_s(n, 1, sizeof(struct nlpcdbrp));
		memcpy(n, &pair, sizeof(struct nlpcdbrp));
		q_add(&opts.q_mod, n);

		if (NLPCDBR_SCAN != pair.record.state)
			continue;
		if ( ! opts.ascend_scan) 
			continue;
		while (uri_ascend(&pair.record.uri))
			proc_html_enqueue(&pair.record.uri, NULL);
	}

	db_query_close(&cookie);

	if (SIGNAL_EXIT & flags)
		return(flags);
	if (SIGNAL_EXIT & (flags |= scan_new()))
		return(flags);
	if (SIGNAL_EXIT & (flags |= scan_mod()))
		return(flags);

	return(flags);
}


static int
scan_seed(int argc, char *argv[])
{
	char		*p;
	int		i, flags;

	assert(NULL != argv);
	for (i = 0; argc-- > 0 && NULL != (p = *argv++); )
		proc_html_link(NULL, NULL, p);

	flags = 0;
	if (SIGNAL_EXIT & (flags |= scan_new()))
		return(flags);

	return(flags);
}


void
xml_struct(void *arg, xmlErrorPtr p)
{
	(void)arg;
	(void)p;
}


void
xml_generic(void *arg, const char *msg, ...)
{
	(void)arg;
	(void)msg;
}


int
main(int argc, char *argv[])
{
	int 		c;
	char		*dir;

	extern char	*__progname;
	extern char 	*optarg;
	extern int 	optind;

	static struct option longopts[] = {
		{ "filter-scheme", 	required_argument, NULL, 1 },
		{ "filter-auth", 	required_argument, NULL, 2 },
		{ "filter-ascend", 	no_argument, NULL, 3 },
		{ "filter-ascend-scan", no_argument, NULL, 4 },
		{ NULL, 0, NULL, 0 }
	};

	dir = NLPC_DEFAULT_DIR;

	memset(&opts, 0, sizeof(struct opt_t));
	opts.cache = NLPC_DEFAULT_FILES;

	q_new(&opts.q_mod);
	q_new(&opts.q_new);

	while (-1 != (c = getopt_long(argc, argv, "b:hvdf:p:s:",
					longopts, NULL)))
		switch (c) {
		case (1):
			dict_free(&opts.scheme);
			dict_fill(&opts.scheme, optarg);
			break;
		case (2):
			dict_free(&opts.auth);
			dict_fill(&opts.auth, optarg);
			break;
		case (3):
			opts.ascend = 1;
			break;
		case (4):
			opts.ascend_scan = 1;
			break;
		case ('d'):
			dbg_lvl++;
			break;
		case ('f'):
			opts.cache = optarg;
			break;
		case ('p'):
			dir = optarg;
			break;
		case ('v'):
			fprintf(stdout, "%s version %s\n", 
					__progname, VERSION);
			return(EXIT_SUCCESS);
		case ('h'):
			/*FALLTHROUGH*/
		default:
			fprintf(stderr, "usage: %s [-dhv] [-f path] "
					"[-p path] [LONG_OPTS] "
					"[uri ...]\n", __progname);
			return(EXIT_FAILURE);
		}
	argc -= optind;
	argv += optind;

	xmlSetStructuredErrorFunc(NULL, xml_struct);
	xmlSetGenericErrorFunc(NULL, xml_generic);

	signal_init(dir, NLPCDBR_VRFY);
	db_init(dir, 0);

	if (SIGNAL_EXIT & scan_seed(argc, argv))
		goto main_out;

	for (;;) {
		if (SIGNAL_EXIT & scan())
			break;
		if (SIGNAL_EXIT & signal_wait(0))
			break;
	}

main_out:

	dict_free(&opts.scheme);
	dict_free(&opts.auth);

	q_free(&opts.q_new);
	q_free(&opts.q_mod);

	db_teardown();
	xmlCleanupParser();
	signal_teardown();

	return(EXIT_SUCCESS);
}

