/* $Id: uri.c,v 1.21 2007-05-24 16:08:55 kristaps Exp $ */
/*
Copyright (c) 2007 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/param.h>

#include <assert.h>
#include <ctype.h>
#include <limits.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "nlpcrawl.h"

char *
uri_topstring(const struct uri *p) 
{
	static char	s[NLPCR_URI_MAXSZ + 10];
	return(uri_topstring_r(p, s));
}


char *
uri_topstring_r(const struct uri *p, char *s)
{
	assert(NULL != p);
	assert(NULL != s);

	memset(s, 0, NLPCR_URI_MAXSZ + 10);
	snprintf(s, NLPCR_URI_MAXSZ + 9, "[%s][%s][%s][%s][%s]",
			p->scheme, p->auth, p->path, p->query,
			p->fragment);
	return(s);
}


char *
uri_tostring(const struct uri *p)
{
	static char	s[NLPCR_URI_MAXSZ + 10];
	return(uri_tostring_r(s, p));
}


char *
uri_tostring_r(char *s, const struct uri *p)
{
	assert(NULL != p);
	assert(NULL != s);
	memset(s, 0, NLPCR_URI_MAXSZ + 10);
	snprintf(s, NLPCR_URI_MAXSZ + 9, "%s%s%s%s%s%s%s%s%s",
			p->scheme, 0 == p->scheme[0] ? "" : ":",
			0 == p->auth[0] ? "" : "//",
			p->auth, p->path,
			0 == p->query[0] ? "" : "?",
			p->query, 0 == p->fragment[0] ? "" : "#",
			p->fragment);
	return(s);
}


/*
 * Merge two paths into one (as per the rfc 3096 sect. 5.2.2
 * reconstruction algorithm).
 */
static void
uri_path_merge(const struct uri *base, const struct uri *rel,
		struct uri *res)
{
	char		*p;

	assert(NULL != base);
	assert(NULL != rel);
	assert(NULL != res);

	if (0 == rel->path[0]) {
		memcpy(res->path, base->path, NLPCR_URI_PATHSZ);
		return;
	}
	p = res->path;
	if (0 != base->path[0]) {
		memcpy(p, base->path, NLPCR_URI_PATHSZ);
		p += strlen(base->path);
	} else
		*p++ = '/';
	while (p != res->path && *p != '/')
		*p-- = 0;
	p++;
	memcpy(p, rel->path, NLPCR_URI_PATHSZ - (p - res->path));
}


/*
 * Normalise a string to the lowercase.
 */
static int
uri_case_decode(char *p, size_t sz)
{
	size_t		i;

	assert(NULL != p);
	for (i = 0; i < sz; i++)
		if (isalpha(p[i]))
			p[i] = (char)tolower(p[i]);
	return(1);
}


/*
 * Strip the beginning white-space from a string.  Returns the new start
 * of string pointer.
 */
static const char *
uri_trim(const char *uri, size_t *urisz)
{
	const char 	*p;
	size_t		i;

	assert(0 != *urisz);
	assert(NULL != uri);

	for (i = 0, p = uri; i < *urisz; p++, i++)
		if ( ! isspace(*p))
			break;
	assert(i < *urisz);
	return(p);
}


/*
 * Decode superfluous percent-encodings within a uri segment.
 */
static int
uri_pct_decode(char *p, size_t sz)
{
	char		hex[5], *ep, *c;
	size_t		i;
	long		lval;
	int		ival;

	assert(NULL != p);

	for (i = 0, c = p; i < sz; i++, c++) {
		if ('%' != p[i]) {
			*c = p[i];
			continue;
		}
		if (i + 2 >= sz) {
			dddbg("%s: warn: uri-parse: invalid "
					"pct-encode: %s", 
					__func__, p);
			return(0);
		}
		hex[0] = '0';
		hex[1] = 'x';
		hex[2] = p[++i];
		hex[3] = p[++i];
		hex[4] = 0;

		lval = strtol(hex, &ep, 16);
		if (0 != *ep) {
			dddbg("%s: warn: uri-parse: invalid "
					"hex-encode: %s", 
					__func__, hex);
			return(0);
		}
		assert(LONG_MAX != lval);
		assert(LONG_MIN != lval);
		assert(INT_MAX > lval);
		assert(INT_MIN < lval);

		ival = (int)lval;

		if (isalnum(ival) || ('-' == ival) ||
				('.' == ival) || ('_' == ival) ||
				('~' == ival)) {
			*c = ival;
			continue;
		} 

		*c++ = '%';
		*c++ = isalpha(p[i - 1]) ? toupper(p[i-1]) : p[i-1];
		*c = isalpha(p[i]) ? (char)toupper(p[i]) : p[i];
	}
	*c = 0;
	return(1);
}


/*
 * Parse the scheme element of a uri/absolute uri.
 */
static int
uri_parse_scheme(const char *uri, size_t sz, size_t *pos,
		struct uri *res)
{
	char		*p;
	size_t		i;

	assert(NULL != uri);
	assert(NULL != pos);
	assert(NULL != res);
	assert(0 != sz);
	assert(0 == *pos);

	if ( ! isalpha(uri[*pos])) {
		ddbg("%s: warn: uri-parse-scheme: invalid "
				"at %zd: %s", __func__,
				*pos, uri);
		return(0);
	}

	i = 0;
	p = res->scheme;

	/* Normalisation. */

	*p++ = (char)tolower(uri[(*pos)++]);
	i++;

	for ( ; *pos < sz && i < NLPCR_URI_SCHEMESZ; i++) 
		switch (uri[*pos]) {
		case (':'):
			(*pos)++;
			return(1);
		case ('+'):
			/*FALLTHROUGH*/
		case ('-'):
			/*FALLTHROUGH*/
		case ('.'):
			*p++ = uri[(*pos)++];
			break;
		default:
			/* Normalisation. */
			if (isalpha(uri[*pos])) {
				*p++ = (char)tolower(uri[(*pos)++]);
				break;
			} else if (isdigit(uri[*pos])) {
				*p++ = uri[(*pos)++];
				break;
			}
			dddbg("%s: warn: uri-parse-scheme: "
					"invalid at %zd: %s",
					__func__, *pos, uri);
			return(0);
		}

	dddbg("%s: warn: uri-parse-scheme: too long: %s", 
			__func__, res->scheme);
	return(0);
}


/*
 * Parse the "authority" element of the uri: credentials and host.  This
 * globs userinfo components directly into the auth field.
 */
static int
uri_parse_auth(const char *uri, size_t sz, size_t *pos,
		struct uri *res)
{
	char		*p;
	size_t		i;
	int		has_ident;

	assert(NULL != res);
	assert(NULL != uri);
	assert(NULL != pos);

	has_ident = 0;
	p = res->auth;

	for (i = 0; *pos < sz && i < NLPCR_URI_AUTHSZ - 1; i++)
		switch (uri[*pos]) {
		case ('/'):
			/*FALLTHROUGH*/
		case ('?'):
			/*FALLTHROUGH*/
		case ('#'):
			goto auth_norm;
		case ('@'):
			has_ident = 1;
			/*FALLTHROUGH*/
		default:
			*p++ = uri[(*pos)++];
			break;
		}

	if (NLPCR_URI_AUTHSZ - 1 == i) {
		dddbg("%s: warn: uri-parse-auth: too long: %s", 
				__func__, res->auth);
		return(0);
	}

auth_norm:
	/* Three-pass normalisation. */

	if (0 == i)
		return(1);

	p = res->auth;
	if (has_ident) while ('@' != *p++) 
		/*Skip user-info to the host.*/ ;
	if ( ! uri_case_decode(p, i)) 
		return(0);

	if ( ! uri_pct_decode(res->auth, i))
		return(0);
	i = strlen(res->auth);

	p = res->auth;
	if (has_ident) while ('@' != *p++) 
		/*Skip user-info to the host. */ ;
	if ( ! uri_case_decode(p, i)) 
		return(0);

	return(1);
}


/*
 * Parse the document-fragment element of a uri.
 */
static int
uri_parse_frag(const char *uri, size_t sz, size_t *pos, 
		struct uri *res)
{
	char		*p;
	size_t		i;

	assert(NULL != res);
	assert(NULL != uri);
	assert(NULL != pos);

	p = res->fragment;
	for (i = 0; *pos < sz && i < NLPCR_URI_FRAGSZ - 1; i++) 
		*p++ = uri[(*pos)++];

	if (NLPCR_URI_FRAGSZ - 1 != i)
		return(1);
	dddbg("%s: warn: uri-parse-frag: too long: %s", 
			__func__, res->fragment);
	return(0);
}


/*
 * Parse the query-string base element of a uri.
 */
static int
uri_parse_query(const char *uri, size_t sz, size_t *pos, 
		struct uri *res)
{
	char		*p;
	size_t		i;

	assert(NULL != res);
	assert(NULL != uri);
	assert(NULL != pos);

	p = res->query;
	for (i = 0; *pos < sz && i < NLPCR_URI_QUERYSZ - 1; i++)
		switch (uri[*pos]) {
		case ('#'):
			return(1);
		default:
			*p++ = uri[(*pos)++];
			break;
		}

	if (NLPCR_URI_FRAGSZ - 1 != i)
		return(1);
	dddbg("%s: warn: uri-parse-query: too long: %s", 
			__func__, res->query);
	return(0);
}


/*
 * Normalise a generic-form uri (ftp, http, https, ...), all of which
 * have the same path-, auth-, and prefix-semantics.
 */
static int
uri_norm_generic(struct uri *res, const char *ps)
{
	char		*p, *save;
	const char	*add;

	assert(NULL != ps);
	assert(NULL != res);

	if (0 != res->auth[0]) {
		if (NULL == (p = strchr(res->auth, '@')))
			p = res->auth;
		add = NULL;
		if (NULL != (p = strchr(p, ':'))) {
			for (save = ++p; 0 != *save; save++) {
				if (isdigit(*save))
					continue;
				dddbg("%s: warn: uri-parse-auth: "
					"invalid port: %s", 
					__func__, res->auth);
				return(0);
			}
			if (save == p) 
				add = ps + 1;
		} else 
			add = ps;
		if (NULL != add) {
			if (NLPCR_URI_AUTHSZ - 1 < strlen(res->auth) + 
					strlen(add)) {
				dddbg("%s: warn: uri-parse-auth: too "
					"long: %s", 
					__func__, res->auth);
				return(0);
			}
			strncat(res->auth, add, NLPCR_URI_AUTHSZ - 1);
		}
	}

	if (0 == res->path[0])
		strncat(res->path, "/", NLPCR_URI_AUTHSZ - 1);

	return(1);
}


/*
 * Per-protocol normalisation.  Branches depending on known hierarchic
 * forms ("generic syntax") which we may generally normalise without
 * fear of error (http, https, ftp).
 */
static int
uri_protocol_norm(struct uri *res)
{
	assert(NULL != res);
	if (0 == res->scheme[0])
		return(1);

	if (0 == strcmp(res->scheme, "http"))
		return(uri_norm_generic(res, ":80"));
	if (0 == strcmp(res->scheme, "https"))
		return(uri_norm_generic(res, ":443"));
	if (0 == strcmp(res->scheme, "ftp"))
		return(uri_norm_generic(res, ":20"));

	return(1);
}


/*
 * Parse and normalise a path segment to a uri.
 */
static int
uri_parse_path(const char *uri, size_t sz, size_t *pos, 
		struct uri *res, int norm)
{
	char		*p, *s, buf[NLPCR_URI_PATHSZ], 
			swap[NLPCR_URI_PATHSZ];
	size_t		i, j;

	assert(NULL != res);
	assert(NULL != uri);
	assert(NULL != pos);

	p = res->path;
	for (i = 0; *pos < sz && i < NLPCR_URI_PATHSZ - 1; i++)
		switch (uri[*pos]) {
		case ('?'):
			/*FALLTHROUGH*/
		case ('#'):
			goto path_norm;
		default:
			*p++ = uri[(*pos)++];
			break;
		}

	if (NLPCR_URI_FRAGSZ - 1 == i) {
		dddbg("%s: warn: uri-parse-path: too long: %s", 
				__func__, res->path);
		return(0);
	}

path_norm:
	if (0 == i)
		return(1);

	if ( ! uri_pct_decode(res->path, i))
		return(0);
	i = strlen(res->path);

	if ( ! norm)
		return(1);

	memset(buf, 0, NLPCR_URI_PATHSZ);
	s = buf;
	p = res->path;

	while (0 != *p) {
		for (j = 0; 0 != *p && '/' != *p; j++, p++)
			swap[j] = *p;
		swap[j] = 0;

		if (0 == strcmp(swap, "..")) {
			j = 0;
			if (s != buf) {
				if (--s != buf) do {
					*s-- = 0;
				} while ('/' != *s && s != buf);
				if (*s != '/')
					*s = 0;
				else
					s++;
			}
		} else if (0 == strcmp(swap, ".")) 
			j = 0;

		memcpy(s, swap, j);
		s += j;
		if ('/' == *p) {
			if (s == buf)
				*s++ = '/';
			else if (s != buf && ('/' != *(s - 1)))
				*s++ = '/';
			p++;
		}
	}

	memcpy(res->path, buf, NLPCR_URI_PATHSZ);
	return(1);
}


/*
 * Parse an absolute uri: a mis-nomer, as this doesn't enforce the
 * absolute uri's non-fragment clause.  Instead, this merely parses from
 * the scheme.
 */
static int
uri_parse_from_scheme(const char *uri, size_t urisz, struct uri *res)
{
	size_t		pos;

	assert(NULL != uri);
	assert(NULL != res);

	if (NLPCR_URI_MAXSZ < urisz) {
		ddbg("%s: warn: uri-parse: too long: %s", 
				__func__, uri);
		return(0);
	}

	pos = 0;
	if ( ! uri_parse_scheme(uri, urisz, &pos, res))
		return(0);

	switch (uri[pos]) {
	case ('?'):
		goto abs_query;
	case ('/'):
		break;
	default:
		goto abs_path;
	}

	if ((2 <= urisz - pos) && (('/' == uri[pos]) &&
				('/' == uri[pos + 1]))) {
		pos += 2;
		if ( ! uri_parse_auth(uri, urisz, &pos, res))
			return(0);
	}

abs_path:
	if ( ! uri_parse_path(uri, urisz, &pos, res, 1))
		return(0);

abs_query:
	if ('?' == uri[pos]) {
		pos++;
		if ( ! uri_parse_query(uri, urisz, &pos, res))
			return(0);
	}

	if (pos != urisz) {
		dddbg("%s: warn: uri-parse: leftover tokens: "
				"%s", __func__, uri);
		return(0);
	}

	if ( ! uri_protocol_norm(res))
		return(0);

	return(1);
}


/*
 * Parse a relative reference; or in other words, a uri without a
 * prefixing scheme (other stipulations are colons in the path segment,
 * which we don't enforce).
 */
static int
uri_parse_from_ref(const char *uri, size_t urisz, struct uri *res)
{
	size_t		pos;

	assert(NULL != uri);
	assert(NULL != res);
	
	if (NLPCR_URI_MAXSZ < urisz) {
		dddbg("%s: warn: uri-parse: too long: %s", 
				__func__, uri);
		return(0);
	}

	switch (uri[(pos = 0)]) {
	case ('?'):
		goto rel_query;
	case ('#'):
		goto rel_fragment;
	case ('/'):
		break;
	default:
		goto rel_path;
	}

	if ((2 <= urisz - pos) && (('/' == uri[pos]) &&
			       ('/' == uri[pos + 1]))) {
		pos += 2;
		if ( ! uri_parse_auth(uri, urisz, &pos, res))
			return(0);
	}

rel_path:
	if ( ! uri_parse_path(uri, urisz, &pos, res, 1))
		return(0);

rel_query:
	if ('?' == uri[pos]) {
		pos++;
		if ( ! uri_parse_query(uri, urisz, &pos, res))
			return(0);
	}

rel_fragment:
	if ('#' == uri[pos]) {
		pos++;
		if ( ! uri_parse_frag(uri, urisz, &pos, res))
			return(0);
	}

	assert(pos == urisz);
	if ( ! uri_protocol_norm(res))
		return(0);

	return(1);
}


/*
 * Quick check whether there's a heading scheme.
 */
static int
uri_head_scheme(const char *uri, size_t urisz)
{
	size_t		i;

	assert(NULL != uri);
	assert(0 != urisz);

	if ( ! isalpha(*uri++))
		return(0);
	for (i = 1; i < urisz; i++, uri++) {
		switch (*uri) {
		case (':'):
			return(1);
		case ('+'):
			/*FALLTHROUGH*/
		case ('-'):
			/*FALLTHROUGH*/
		case ('.'):
			break;
		default:
			if ( ! isalnum(*uri))
				return(0);
			break;
		}
	}
	return(0);
}


int
uri_parse(const char *uri, size_t urisz, struct uri *res)
{
	assert(NULL != uri);
	assert(NULL != res);

	memset(res, 0, sizeof(struct uri));

	if (0 == urisz)
		urisz = strlen(uri);
	if (0 == urisz)
		return(1);
		/* This was a valid relative-reference uri. */

	if (NULL == (uri = uri_trim(uri, &urisz)))
		return(0);

	if (uri_head_scheme(uri, urisz)) 
		return(uri_parse_from_scheme(uri, urisz, res));
	return(uri_parse_from_ref(uri, urisz, res));
}


int
uri_is_abs(const struct uri *p)
{
	assert(NULL != p);
	return(0 != p->scheme[0] && 0 == p->fragment[0]);
}


int
uri_ref_resolve(const struct uri *base, const struct uri *rel,
		struct uri *res)
{
	assert(NULL != base);
	assert(NULL != rel);
	assert(NULL != res);
	assert(uri_is_abs(base));

	memset(res, 0, sizeof(struct uri));

	if (0 != rel->scheme[0]) {
		memcpy(res->scheme, rel->scheme, NLPCR_URI_SCHEMESZ);
		memcpy(res->auth, rel->auth, NLPCR_URI_AUTHSZ);
		memcpy(res->path, rel->path, NLPCR_URI_PATHSZ);
		memcpy(res->query, rel->query, NLPCR_URI_QUERYSZ);
	} else {
		if (0 != rel->auth[0]) {
			memcpy(res->auth, rel->auth, NLPCR_URI_AUTHSZ);
			memcpy(res->path, rel->path, NLPCR_URI_PATHSZ);
			memcpy(res->query, rel->query, 
					NLPCR_URI_QUERYSZ);
		} else {
			if (0 == rel->path[0]) {
				memcpy(res->path, base->path, 
						NLPCR_URI_PATHSZ);
				memcpy(res->query, 0 != rel->query[0]?
					rel->query : base->query,
					NLPCR_URI_QUERYSZ);
			} else {
				if ('/' == rel->path[0])
					memcpy(res->path, rel->path,
						NLPCR_URI_PATHSZ);
				else uri_path_merge(base, rel, res);
				memcpy(res->query, rel->query,
						NLPCR_URI_QUERYSZ);
			}
			memcpy(res->auth, base->auth, 
					NLPCR_URI_AUTHSZ);
		}
		memcpy(res->scheme, base->scheme, NLPCR_URI_SCHEMESZ);
	}
	memcpy(res->fragment, rel->fragment, NLPCR_URI_SCHEMESZ);

	if (0 != res->scheme[NLPCR_URI_SCHEMESZ - 1]) {
		dddbg("%s: fail: ref. resolution: scheme too long",
				__func__);
		return(0);
	}
	if (0 != res->auth[NLPCR_URI_AUTHSZ - 1]) {
		dddbg("%s: fail: ref. resolution: auth too long",
				__func__);
		return(0);
	}
	if (0 != res->path[NLPCR_URI_PATHSZ - 1]) {
		dddbg("%s: fail: ref. resolution: path too long",
				__func__);
		return(0);
	}
	if (0 != res->query[NLPCR_URI_QUERYSZ - 1]) {
		dddbg("%s: fail: ref. resolution: query too long",
				__func__);
		return(0);
	}
	if (0 != res->fragment[NLPCR_URI_FRAGSZ - 1]) {
		dddbg("%s: fail: ref. resolution: fragment too long",
				__func__);
		return(0);
	}

	return(1);
}


int
uri_ascend(struct uri *addr)
{
	char		*p;
	size_t		sz;

	assert(NULL != addr);

	if (0 != addr->fragment[0]) {
		memset(addr->fragment, 0, NLPCR_URI_FRAGSZ);
		return(1);
	}

	if (0 != addr->query[0]) {
		memset(addr->query, 0, NLPCR_URI_QUERYSZ);
		return(1);
	}

	if (0 == (sz = strlen(addr->path)))
		return(0);
	if (1 == sz) {
		/* FIXME: uri should be protocol-normalised. */
		if ('/' == addr->path[0])
			return(0);
		memset(addr->path, 0, NLPCR_URI_PATHSZ);
		return(1);
	} else if ('/' == addr->path[sz - 1]) {
		addr->path[sz - 1] = 0;
		sz--;
	}

	p = addr->path + sz - 1;
	while (p != addr->path && '/' != *p)
		*p-- = 0;
	return(1);
}

