/* $Id: vrfy.c,v 1.76 2007-05-25 10:31:03 kristaps Exp $ */
/*
Copyright (c) 2007 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/param.h>

#include <assert.h>
#include <err.h>
#include <ctype.h>
#include <errno.h>
#include <getopt.h>
#include <inttypes.h>
#include <limits.h>
#include <locale.h>
#include <signal.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <wctype.h>

#include <libxml/HTMLparser.h>
#include <libxml/HTMLtree.h>

#include "nlpcrawl.h"


/* Default percentage of required words to "pass". */
#define VRFY_DICT_PCT	(10)

/* Default charset. */
#define VRFY_DEF_CHARSET "iso-8859-1"


static struct opt_t {
	void		*queue;		/* In-memory write queue.    */
	char		*cache;		/* Cache-file directory.     */
	void		*enc;		/* Valid code htable (opt).  */
	void		*dict;		/* Valid word htable (opt).  */
	unsigned short	dictpct;	/* %-needed in dict.         */
} opts;


static struct stat_t {
	uint64_t	count;		/* Total records mod/add.    */
	uint64_t	starttime;	/* Start of run-time.        */
	uint64_t	count_unreach;	/* Records unreachable.      */
	uint64_t	count_vrfy;	/* Records verified.         */
	uint64_t	enc_check;	/* Encoding-checked.         */
	uint64_t	enc_fail;	/* Encoding-check-failed.    */
	uint64_t	wc_check;	/* Word-count checked.       */
	uint64_t	wc_fail;	/* Word-count match failed.  */
	uint64_t	wc_total;	/* Word-count total.         */
	uint64_t	wc_matches;	/* Word-count matches total. */
} stats;


/*
 * Statistics on word count carried around by an active parse.  This is
 * only used when dictionary-matching has been enabled.
 */
struct wordstat {
	size_t		total;		/* Total words checked.      */
	size_t		matches;	/* Matches in htable.        */
};


static int
proc_dict(const struct wordstat *st)
{
	unsigned short	pct;
	double		dec;

	assert(NULL != st);
	if (0 == st->matches)
		return(0);

	dec = (double)st->matches / (double)st->total;
	assert(1.0 >= dec);
	pct = (unsigned short)(dec * 100);

	ddbg("dict-match: %zd, %zd, %u%%", st->matches, st->total, pct);
	return (pct > opts.dictpct);
}


/*
 * Convert a lower-case word to single-byte format and attempt to match 
 * it in the hashtable.  Returns 0 on failure, 1 on success.
 */
static int
proc_word(struct wordstat *st, const struct nlpcdbr *rec, 
		wchar_t word[HASH_FILE_WORD])
{
	char		raw[HASH_FILE_WORD];
	size_t		sz;

	assert(NULL != st);

	if (L'\0' == word[0])
		return(1);
	if (NULL == opts.dict) 
		return(1);

	sz = wcstombs(raw, word, HASH_FILE_WORD);
	if (HASH_FILE_WORD == sz) {
		ddbg("fail: wcstombs overrun: %s", 
				uri_topstring(&rec->uri));
		return(0);
	} else if ((size_t)-1 == sz) {
		int errn = errno;
		ddbg("fail: wcstombs: %s, %s", strerror(errn),
				uri_topstring(&rec->uri));
		return(0);
	}

	st->total++;
	if (dict_has(&opts.dict, raw, strlen(raw)))
		st->matches++;

	return(1);
}


/*
 * This needs the most work.  Process a chunk of text.  The chunk will
 * be in UTF-8 format as an xmlChar (unsigned char).  Returns 0 on
 * failure, 1 on success.
 */
static int
proc_node(struct wordstat *st, 
		const struct nlpcdbr *rec, xmlNodePtr node)
{
	xmlNodePtr 	c;
	xmlChar		*data;
	size_t		i, bufsz, sz, ssz;
	wchar_t		*buf, *bufp, str[HASH_FILE_WORD], p;
	int		rc;

	assert(NULL != st);

	if (NULL == node)
		return(1);

	bufp = buf = NULL;
	bufsz = 0;
	rc = 1;

	for (c = node; NULL != c; c = c->next) {
		if ( ! proc_node(st, rec, c->children)) {
			rc = 0;
			goto proc_node_out;
		}
		if (c->type != XML_TEXT_NODE)
			continue;
		if (NULL == (data = c->content))
			continue;

		sz = sizeof(wchar_t) * (strlen((char *)data) + 1);
		if (sz > bufsz) {
			if (NULL == (buf = realloc(buf, sz)))
				die("fail: allocate");
			bufsz = sz;
		}

		assert(NULL != buf);
		assert(NULL != data);

		ssz = mbstowcs(buf, (char *)data, sz);
		if (sz == ssz) {
			ddbg("fail: mbstowcs overrun: %s", 
					uri_topstring(&rec->uri));
			rc = 0;
			goto proc_node_out;
		} else if ((size_t)-1 == ssz) {
			int errn = errno;
			ddbg("fail: mbstowcs: %s, %s", strerror(errn),
					uri_topstring(&rec->uri));
			rc = 0;
			goto proc_node_out;
		}

		for (bufp = buf, i = 0; ; ) {
			if (i > HASH_FILE_WORD - 1) {
				ddbg("fail: word overrun: %zd, %s",
					i, uri_topstring(&rec->uri));
				rc = 0;
				goto proc_node_out;
			}
			p = *bufp++;
			if (0 == p || L'\0' == p) {
				str[i] = L'\0';
				if (proc_word(st, rec, str))
					break;
				rc = 0;
				goto proc_node_out;
			}
			if (iswspace(p)) {
				str[i] = L'\0';
				if (proc_word(st, rec, str)) {
					i = 0;
					continue;
				}
				rc = 0;
				goto proc_node_out;
			}

			/* Convert to lowercase. */
			str[i++] = iswalpha(p) ? 
				(wchar_t)towlower(p) : p;
		}
	}

proc_node_out:

	if (NULL != buf)
		free(buf);
	return(rc);
}


/*
 * Return 1 on success, 0 on skip, -1 on fail.
 */
static int
proc_tail(struct nlpcdbrp *pair, xmlDocPtr doc)
{
	struct nlpcdbr	*rec;
	struct wordstat	st;
	xmlNodePtr	p;

	assert(NULL != pair);
	assert(NULL != doc);

	memset(&st, 0, sizeof(struct wordstat));
	rec = &pair->record;

	/*
	 * Validate the charset.  This returns "true" if no database has
	 * been set.
	 */

	if ( ! dict_has(&opts.enc, rec->charset, 
				strlen(rec->charset))) {
		dbg("reject: illegal charset: %s, %s", rec->charset,
				uri_topstring(&rec->uri));
		xmlFreeDoc(doc);
		rec->code = NLPCDBR_CHARSET;
		stats.count++;
		stats.count_unreach++;
		stats.enc_check++;
		stats.enc_fail++;
		return(-1);
	}

	/*
	 * Now parse the document and seek out text sections to compare
	 * against the dictionary.
	 */

	if (NULL == (p = xmlDocGetRootElement(doc))) {
		dbg("reject: failed acquire root: %s", 
				uri_topstring(&rec->uri));
		xmlFreeDoc(doc);
		rec->code = NLPCDBR_SYNTAX;
		stats.count++;
		stats.count_unreach++;
		stats.enc_check++;
		return(-1);
	} else if ( ! proc_node(&st, rec, p)) {
		dbg("reject: failed node parse: %s",
				uri_topstring(&rec->uri));
		xmlFreeDoc(doc);
		rec->code = NLPCDBR_SYNTAX;
		stats.count++;
		stats.count_unreach++;
		stats.enc_check++;
		return(-1);
	}

	/*
	 * Check against the dictionary and record our outcome
	 * appropriately. 
	 */

	stats.count++;
	stats.enc_check++;
	stats.wc_check++;
	stats.wc_total += st.total;
	stats.wc_matches += st.matches;
	if ( ! proc_dict(&st)) {
		stats.count_unreach++;
		stats.wc_fail++;
		xmlFreeDoc(doc);
		if (0 == st.total) {
			dbg("reject: zero word-count: %s",
					uri_topstring(&rec->uri));
			return(-1);
		}
		dbg("reject: failed dict: %d%% (%u%%), %s",
				(int)(((double)st.matches / 
				       (double)st.total) * 100),
				opts.dictpct,
				uri_topstring(&rec->uri));
		return(-1);
	}
	stats.count_vrfy++;

	xmlFreeDoc(doc);
	return(1);
}


/*
 * Returns 1 on success, 0 on skip, -1 on failure.
 */
static int
proc_xml(struct nlpcdbrp *pair)
{
	xmlDocPtr 	doc;
	struct nlpcdbr	*rec;
	int		i;
	char		*path, *charset;
	const char	*enc;

	assert(NULL != pair);
	rec = &pair->record;

	path = nlpcdbr_hash2path(opts.cache, pair->hash);
	charset = 0 == rec->charset[0] ? NULL : rec->charset;

	/* 
	 * First parse the file into a DOM tree.
	 */

	if (NULL == (doc = xmlReadFile(path, charset,
					XML_PARSE_NONET))) {
		dbg("reject: failed document parse: %s", 
				uri_topstring(&rec->uri));
		rec->code = NLPCDBR_SYNTAX;
		stats.count++;
		stats.count_unreach++;
		return(-1);
	}

	/*
	 * Let libxml determine the encoding for us.  If there's no
	 * encoding, use libxml's, else use the default.  Make sure
	 * that's lower-cased.
	 */

	if (0 == rec->charset[0]) {
		if (NULL == (enc = (const char *)doc->encoding))
			enc = VRFY_DEF_CHARSET;
		/* This can be zero-length ("charset="). */
		if (0 == enc[0])
			enc = VRFY_DEF_CHARSET;
		strncpy(rec->charset, enc, NLPCR_CHARSETLEN);
		assert(0 == rec->charset[NLPCR_CHARSETLEN - 1]);
		for (i = 0; i < NLPCR_CHARSETLEN; i++)
			if (isalpha(rec->charset[i]))
				rec->charset[i] = 
					tolower(rec->charset[i]);
		ddbg("charset-shift: %s, %s", rec->charset, 
				uri_topstring(&rec->uri));
	}

	return(proc_tail(pair, doc));
}


/*
 * Returns 1 on success, -1 on failure, 0 on skip.
 */
static int
proc_html(struct nlpcdbrp *pair)
{
	htmlDocPtr 	doc;
	struct nlpcdbr	*rec;
	int		i;
	char		*path, *charset;
	const char	*enc;

	assert(NULL != pair);
	rec = &pair->record;

	path = nlpcdbr_hash2path(opts.cache, pair->hash);
	charset = 0 == rec->charset[0] ? NULL : rec->charset;

	/* 
	 * First parse the file into a DOM tree.
	 */

	if (NULL == (doc = htmlReadFile(path, charset,
					HTML_PARSE_NONET))) {
		dbg("reject: failed document parse: %s", 
				uri_topstring(&rec->uri));
		rec->code = NLPCDBR_SYNTAX;
		stats.count++;
		stats.count_unreach++;
		return(-1);
	}

	/*
	 * Let libxml determine the encoding for us.  If there's no
	 * encoding, use libxml's, else use the default.  Make sure
	 * that's lower-cased.
	 */

	if (0 == rec->charset[0]) {
		enc = (const char *)htmlGetMetaEncoding(doc);
		if (NULL == enc)
			enc = VRFY_DEF_CHARSET;
		/* This can be zero-length ("charset="). */
		if (0 == enc[0])
			enc = VRFY_DEF_CHARSET;
		strncpy(rec->charset, enc, NLPCR_CHARSETLEN);
		assert(0 == rec->charset[NLPCR_CHARSETLEN - 1]);
		for (i = 0; i < NLPCR_CHARSETLEN; i++)
			if (isalpha(rec->charset[i]))
				rec->charset[i] = 
					tolower(rec->charset[i]);
		ddbg("charset-shift: %s, %s", rec->charset, 
				uri_topstring(&rec->uri));
	}

	return(proc_tail(pair, (xmlDocPtr)doc));
}


/* 
 * Return 0 on skip (no change to record), -1 on fail (record moves to
 * UNREACH), or 1 (record moves to VRFY).  
 */
static int
proc(struct nlpcdbrp *pair)
{
	int		rc;
	struct nlpcdbr	*rec;

	assert(NULL != pair);
	rec = &pair->record;

	switch (rec->type) {
	case (NLPCDBR_TYPE_TXT_HTML):
		if (1 != (rc = proc_html(pair)))
			return(rc);
		break;
	case (NLPCDBR_TYPE_TXT_XML):
		/*FALLTHROUGH*/
	case (NLPCDBR_TYPE_APP_XML):
		/*FALLTHROUGH*/
	case (NLPCDBR_TYPE_APP_XHTML):
		if (1 != (rc = proc_xml(pair)))
			return(rc);
		break;
	default:
		ddbg("warn: unsupported content-type: %s, %s",
				nlpcdbr_type2str(rec->type), 
				uri_topstring(&rec->uri));
		return(0);
	}

	dbg("accept: %s", uri_topstring(&rec->uri));
	return(1);
}


static int
scan_mod(void)
{
	int		vrfy, unreach, flags;
	struct uri	*addr;
	struct nlpcdbrp	*n;

	if (q_empty(&opts.queue))
		return(0);

	for (vrfy = 0, unreach = 0, flags = 0; ;) {
		n = (struct nlpcdbrp *)q_rem(&opts.queue);
		if (NULL == n)
			break;
		addr = &n->record.uri;

		if (SIGNAL_EXIT & (flags |= signal_test()))
			break;

		if ( ! db_put(n))
			diex("fail-modify: %s",uri_topstring(addr));

		switch (n->record.state) {
		case (NLPCDBR_UNREACH):
			unreach++;
			break;
		case (NLPCDBR_VRFY):
			vrfy++;
			break;
		default:
			abort();
		}
		free(n);
	}
	if (vrfy > 0)
		signal_next(NLPCDBR_VRFY);
	if (unreach > 0)
		signal_next(NLPCDBR_UNREACH);

	return(flags);
}


static int
scan(void)
{
	int		rc, flags;
	struct db_query	query;
	struct nlpcdbrp pair, *n;
	void		*cookie;

	memset(&query, 0, sizeof(struct db_query));

	query.field = DB_FIELD_STATE;
	query.state = NLPCDBR_FRESH;
	query.window = -1;

	db_window_fill(&query);
	db_query_begin(&cookie);

	for (flags = 0; ;) {
		if (SIGNAL_EXIT & (flags |= signal_test()))
			break;
		if (1 != db_query_next(&cookie, &query, &pair))
			break;
		if (0 == (rc = proc(&pair)))
			continue;

		pair.record.atime = pair.record.mtime = time_uint64();
		pair.record.state = -1 == rc ?
			NLPCDBR_UNREACH : NLPCDBR_VRFY;

		calloc_s(n, 1, sizeof(struct nlpcdbrp));
		memcpy(n, &pair, sizeof(struct nlpcdbrp));

		q_add(&opts.queue, n);
	}

	db_query_close(&cookie);

	if (SIGNAL_EXIT & flags)
		return(flags);
	if (SIGNAL_EXIT & (flags |= scan_mod()))
		return(flags);

	return(flags);
}


void
xml_struct(void *arg, xmlErrorPtr p)
{
	(void)arg;
	if ('\n' == p->message[strlen(p->message) - 1])
		p->message[strlen(p->message) - 1] = 0;
	dddbg("warn: %s, %s:%d", p->message, NULL == p->file ? 
			"(none)" : p->file, p->line);
}


void
xml_generic(void *arg, const char *fmt, ...)
{
	(void)arg;
	va_list ap;
	va_start(ap, fmt);
	vdbg(3, fmt, ap);
	va_end(ap);
}


int
main(int argc, char *argv[])
{
	extern char	*__progname, *optarg;
	extern int 	optind;

	int 		c, t;
	char		*dir, *dict;

	static struct option longopts[] = {
		{ "filter-charset", 	required_argument, NULL, 1 },
		{ "filter-dict", 	required_argument, NULL, 2 },
		{ "filter-dict-pct", 	required_argument, NULL, 3 },
		{ NULL, 0, NULL, 0 }
	};

	if (NULL == setlocale(LC_CTYPE, "en_GB.UTF-8"))
		die("fail: setting locale");

	dir = NLPC_DEFAULT_DIR;
	dict = NULL;

	memset(&stats, 0, sizeof(struct stat_t));
	stats.starttime = time_uint64();

	memset(&opts, 0, sizeof(struct opt_t));
	opts.dictpct = VRFY_DICT_PCT;
	opts.cache = NLPC_DEFAULT_FILES;
	q_new(&opts.queue);

	while (-1 != (c = getopt_long(argc, argv, "f:vhdp:",
					longopts, NULL)))
		switch (c) {
		case (1):
			dict_free(&opts.enc);
			dict_fill(&opts.enc, optarg);
			break;
		case (2):
			dict_free(&opts.dict);
			dict_file(&opts.dict, optarg);
			break;
		case (3):
			if ( ! parse_int(optarg, &t, 100, 1))
				diex("invalid value: --dict-pct %s", 
						optarg);
			opts.dictpct = (unsigned short)t;
			break;
		case ('d'):
			dbg_lvl++;
			break;
		case ('f'):
			opts.cache = optarg;
			break;
		case ('p'):
			dir = optarg;
			break;
		case ('v'):
			fprintf(stdout, "%s version %s\n", 
					__progname, VERSION);
			return(EXIT_SUCCESS);
		case ('h'):
			/*FALLTHROUGH*/
		default:
			fprintf(stderr, "usage: %s [-dhv] [-f path] "
					"[-p path] [LONG_OPTS]\n", 
					__progname);
			return(EXIT_FAILURE);
		}
	argc -= optind;
	argv += optind;

	xmlSetStructuredErrorFunc(NULL, xml_struct);
	xmlSetGenericErrorFunc(NULL, xml_generic);

	signal_init(dir, NLPCDBR_FRESH);
	db_init(dir, 0);

	for (;;) {
		if (SIGNAL_EXIT & scan())
			break;
		if (SIGNAL_EXIT & signal_wait(0))
			break;
	}

	dict_free(&opts.enc);
	dict_free(&opts.dict);
	q_free(&opts.queue);

	db_teardown();
	xmlCleanupParser();
	signal_teardown();

	return(EXIT_SUCCESS);
}

